'''Numpy/HDF5 data types shared among several WESTPA tools'''

import numpy as np

# Pick up a few data types from the WEST core if possible
try:
    from westpa.core.data_manager import n_iter_dtype, seg_id_dtype, weight_dtype
except ImportError:
    n_iter_dtype = np.uint32
    seg_id_dtype = np.int64
    weight_dtype = np.float64

# A quantity averaged over iterations
iter_block_ci_dtype = np.dtype(
    [
        ('iter_start', n_iter_dtype),
        ('iter_stop', n_iter_dtype),
        ('expected', np.float64),
        ('ci_lbound', np.float64),
        ('ci_ubound', np.float64),
        ('sterr', np.float64),
        ('corr_len', n_iter_dtype),
    ]
)

# A quantity to store event duration distribution stuff.
# Comes from the old w_kinetics.

ed_list_dtype = np.dtype(
    [('istate', np.uint16), ('fstate', np.uint16), ('duration', np.float64), ('weight', np.float64), ('seg_id', seg_id_dtype)]
)



import logging


import h5py

log = logging.getLogger(__name__)


class WESTAnalysisTool:
    def __init__(self):
        super().__init__()
        # Whether a west.cfg is required to run a program based on this tool
        self.config_required = False

        # Analysis HDF5 filename and object
        self.anal_h5name = None
        self.anal_h5file = None

        # Whether this is being used in a brute analysis
        self.bf_mode = False

        # A way to override some arguments on a per-mixin basis without having to subclass
        # (messy, but it doesn't seem crucial enough so far to make it cleaner)
        self.include_args = {}

    def add_args(self, parser, upcall=True):
        '''Add arguments to a parser common to all analyses of this type.'''
        if upcall:
            try:
                upfunc = super().add_args
            except AttributeError:
                pass
            else:
                upfunc(parser)

        group = parser.add_argument_group('general analysis options')
        group.add_argument(
            '-A',
            '--analysis-file',
            dest='anal_h5name',
            metavar='H5FILE',
            default='analysis.h5',
            help='Store intermediate and final results in H5FILE (default: %(default)s).',
        )

    def process_args(self, args, upcall=True):
        self.anal_h5name = args.anal_h5name

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def open_analysis_backing(self):
        if self.anal_h5file is None:
            self.anal_h5file = h5py.File(self.anal_h5name)

    def close_analysis_backing(self):
        try:
            self.anal_h5file.close()
            self.anal_h5file = None
        except AttributeError:
            pass

    def require_analysis_group(self, groupname, replace=False):
        self.open_analysis_backing()
        if replace:
            try:
                del self.anal_h5file[groupname]
            except KeyError:
                pass
        return self.anal_h5file.require_group(groupname)



import logging
import operator

import numpy as np

import westpa
from westpa.core.yamlcfg import check_bool
from westpa.core.kinetics import RateAverager
from westpa.westext.weed.ProbAdjustEquil import probAdjustEquil
from westpa.core._rc import bins_from_yaml_dict

EPS = np.finfo(np.float64).eps

log = logging.getLogger(__name__)


class WEEDDriver:
    def __init__(self, sim_manager, plugin_config):
        if not sim_manager.work_manager.is_master:
            return

        self.sim_manager = sim_manager
        self.data_manager = sim_manager.data_manager
        self.system = sim_manager.system
        self.work_manager = sim_manager.work_manager

        self.do_reweight = check_bool(
            plugin_config.get('do_equilibrium_reweighting', False) or plugin_config.get('do_reweighting', False)
        )
        self.windowsize = 0.5
        self.windowtype = 'fraction'

        windowsize = plugin_config.get('window_size')
        if windowsize is not None:
            if isinstance(windowsize, float):
                self.windowsize = windowsize
                self.windowtype = 'fraction'
                if self.windowsize <= 0 or self.windowsize > 1:
                    raise ValueError('WEED parameter error -- fractional window size must be in (0,1]')
            elif isinstance(windowsize, int):
                self.windowsize = int(windowsize)
                self.windowtype = 'fixed'
            else:
                raise ValueError('WEED parameter error -- invalid window size {!r}'.format(windowsize))
        log.info('using window size of {!r} ({})'.format(self.windowsize, self.windowtype))

        self.max_windowsize = plugin_config.get('max_window_size')
        if self.max_windowsize is not None:
            log.info('Using max windowsize of {:d}'.format(self.max_windowsize))

        self.reweight_period = plugin_config.get('reweight_period', 0)
        self.priority = plugin_config.get('priority', 0)

        self.rate_calc_queue_size = plugin_config.get('rate_calc_queue_size', 1)
        self.rate_calc_n_blocks = plugin_config.get('rate_calc_n_blocks', 1)

        bin_obj = plugin_config.get('bins', None)
        if isinstance(bin_obj, dict):
            bin_obj = bins_from_yaml_dict(bin_obj)
        self.bin_mapper = bin_obj

        if self.do_reweight:
            sim_manager.register_callback(sim_manager.prepare_new_iteration, self.prepare_new_iteration, self.priority)

    def get_rates(self, n_iter, mapper):
        '''Get rates and associated uncertainties as of n_iter, according to the window size the user
        has selected (self.windowsize)'''

        if self.windowtype == 'fraction':
            if self.max_windowsize is not None:
                eff_windowsize = min(self.max_windowsize, int(n_iter * self.windowsize))
            else:
                eff_windowsize = int(n_iter * self.windowsize)
        else:  # self.windowtype == 'fixed':
            eff_windowsize = min(n_iter, self.windowsize or 0)

        averager = RateAverager(mapper, self.system, self.data_manager, self.work_manager)
        averager.calculate(max(1, n_iter - eff_windowsize), n_iter + 1, self.rate_calc_n_blocks, self.rate_calc_queue_size)
        self.eff_windowsize = eff_windowsize

        return averager

    def prepare_new_iteration(self):
        n_iter = self.sim_manager.n_iter
        we_driver = self.sim_manager.we_driver

        if we_driver.target_states and self.do_reweight:
            log.warning('equilibrium reweighting requested but target states (sinks) present; reweighting disabled')
            return

        if not self.do_reweight:
            # Reweighting not requested
            log.debug('Equilibrium reweighting not enabled')
            return

        with self.data_manager.lock:
            weed_global_group = self.data_manager.we_h5file.require_group('weed')
            reweighting_history_dataset = weed_global_group.require_dataset(
                'reweighting_history', (1,), maxshape=(None,), dtype=int
            )
            last_reweighting = int(weed_global_group.attrs.get('last_reweighting', 0))
            if last_reweighting > n_iter:
                last_reweighting = n_iter - 1
                reweighting_history = reweighting_history_dataset[:]
                reweighting_history = reweighting_history[reweighting_history < n_iter]
                reweighting_history_dataset.resize((reweighting_history.size), axis=0)

        if n_iter - last_reweighting < self.reweight_period:
            # Not time to reweight yet
            log.debug('not reweighting')
            return
        else:
            log.debug('reweighting')

        if self.bin_mapper is None:
            mapper = we_driver.bin_mapper
            bins = we_driver.next_iter_binning
            westpa.rc.pstatus('\nReweighting using the simulation bin mapper:\n{}'.format(mapper))
        else:
            mapper = self.bin_mapper
            bins = mapper.construct_bins()

            segments = [s for s in we_driver.next_iter_segments]
            pcoords = self.system.new_pcoord_array(len(segments))
            for iseg, segment in enumerate(segments):
                pcoords[iseg] = segment.pcoord[0]
            assignments = mapper.assign(pcoords)
            for segment, assignment in zip(segments, assignments):
                bins[assignment].add(segment)

            westpa.rc.pstatus('\nReweighting using a different bin mapper than simulation:\n{}'.format(mapper))

        n_bins = len(bins)

        # Create storage for ourselves
        with self.data_manager.lock:
            iter_group = self.data_manager.get_iter_group(n_iter)
            try:
                del iter_group['weed']
            except KeyError:
                pass

            weed_iter_group = iter_group.create_group('weed')
            avg_populations_ds = weed_iter_group.create_dataset('avg_populations', shape=(n_bins,), dtype=np.float64)
            unc_populations_ds = weed_iter_group.create_dataset('unc_populations', shape=(n_bins,), dtype=np.float64)
            avg_flux_ds = weed_iter_group.create_dataset('avg_fluxes', shape=(n_bins, n_bins), dtype=np.float64)
            unc_flux_ds = weed_iter_group.create_dataset('unc_fluxes', shape=(n_bins, n_bins), dtype=np.float64)
            avg_rates_ds = weed_iter_group.create_dataset('avg_rates', shape=(n_bins, n_bins), dtype=np.float64)
            unc_rates_ds = weed_iter_group.create_dataset('unc_rates', shape=(n_bins, n_bins), dtype=np.float64)

        averager = self.get_rates(n_iter, mapper)

        with self.data_manager.flushing_lock():
            avg_populations_ds[...] = averager.average_populations
            unc_populations_ds[...] = averager.stderr_populations
            avg_flux_ds[...] = averager.average_flux
            unc_flux_ds[...] = averager.stderr_flux
            avg_rates_ds[...] = averager.average_rate
            unc_rates_ds[...] = averager.stderr_rate

            binprobs = np.fromiter(map(operator.attrgetter('weight'), bins), dtype=np.float64, count=n_bins)
            orig_binprobs = binprobs.copy()

        westpa.rc.pstatus('Calculating equilibrium reweighting using window size of {:d}'.format(self.eff_windowsize))
        westpa.rc.pstatus('\nBin probabilities prior to reweighting:\n{!s}'.format(binprobs))
        westpa.rc.pflush()

        probAdjustEquil(binprobs, averager.average_rate, averager.stderr_rate)

        # Check to see if reweighting has set non-zero bins to zero probability (should never happen)
        assert (~((orig_binprobs > 0) & (binprobs == 0))).all(), 'populated bin reweighted to zero probability'

        # Check to see if reweighting has set zero bins to nonzero probability (may happen)
        z2nz_mask = (orig_binprobs == 0) & (binprobs > 0)
        if (z2nz_mask).any():
            westpa.rc.pstatus('Reweighting would assign nonzero probability to an empty bin; not reweighting this iteration.')
            westpa.rc.pstatus('Empty bins assigned nonzero probability: {!s}.'.format(np.array_str(np.arange(n_bins)[z2nz_mask])))
        else:
            westpa.rc.pstatus('\nBin populations after reweighting:\n{!s}'.format(binprobs))
            for bin, newprob in zip(bins, binprobs):
                bin.reweight(newprob)

            reweighting_history_dataset.resize((reweighting_history_dataset.shape[0] + 1), axis=0)
            reweighting_history_dataset[-1] = n_iter
            weed_global_group.attrs['last_reweighting'] = n_iter

        assert (
            abs(1 - np.fromiter(map(operator.attrgetter('weight'), bins), dtype=np.float64, count=n_bins).sum())
            < EPS * np.fromiter(map(len, bins), dtype=int, count=n_bins).sum()
        )

        westpa.rc.pflush()



import logging

from westpa.tools import WESTParallelTool, WESTDataReader, IterRangeSelection, ProgressIndicatorComponent
import westpa
from westpa.core.extloader import get_object

log = logging.getLogger('w_crawl')


class WESTPACrawler:
    '''Base class for general crawling execution. This class
    only exists on the master.'''

    def initialize(self, iter_start, iter_stop):
        '''Initialize this crawling process.'''
        pass

    def finalize(self):
        '''Finalize this crawling process.'''
        pass

    def process_iter_result(self, n_iter, result):
        '''Process the result of a per-iteration task.'''
        pass


def _remote_task(n_iter, taskfn):
    data_manager = westpa.rc.get_data_manager()  # gaahhh...globals
    data_manager.open_backing(mode='r')
    return n_iter, taskfn(n_iter, data_manager.get_iter_group(n_iter))


class WCrawl(WESTParallelTool):
    prog = 'w_crawl'
    description = '''\
Crawl a weighted ensemble dataset, executing a function for each iteration.
This can be used for postprocessing of trajectories, cleanup of datasets,
or anything else that can be expressed as "do X for iteration N, then do
something with the result". Tasks are parallelized by iteration, and
no guarantees are made about evaluation order.


-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------

'''

    def __init__(self):
        super().__init__()

        # These are used throughout
        self.progress = ProgressIndicatorComponent()
        self.data_reader = WESTDataReader()
        self.iter_range = IterRangeSelection(self.data_reader)

        self.crawler = None
        self.task_callable = None

    def add_args(self, parser):
        self.data_reader.add_args(parser)
        self.iter_range.add_args(parser)

        tgroup = parser.add_argument_group('task options')
        tgroup.add_argument(
            '-c',
            '--crawler-instance',
            help='''Use CRAWLER_INSTANCE (specified as module.instance) as an instance of
                            WESTPACrawler to coordinate the calculation. Required only if initialization,
                            finalization, or task result processing is required.''',
        )
        tgroup.add_argument(
            'task_callable',
            help='''Run TASK_CALLABLE (specified as module.function) on each iteration.
                            Required.''',
        )
        self.progress.add_args(parser)

    def process_args(self, args):
        self.progress.process_args(args)
        self.data_reader.process_args(args)
        with self.data_reader:
            self.iter_range.process_args(args)

        self.task_callable = get_object(args.task_callable, path=['.'])
        if args.crawler_instance is not None:
            self.crawler = get_object(args.crawler_instance, path=['.'])
        else:
            self.crawler = WESTPACrawler()

    def go(self):
        iter_start = self.iter_range.iter_start
        iter_stop = self.iter_range.iter_stop
        iter_count = iter_stop - iter_start
        self.data_reader.open('r')
        pi = self.progress.indicator
        with pi:
            pi.operation = 'Initializing'
            self.crawler.initialize(iter_start, iter_stop)

            try:
                pi.new_operation('Dispatching tasks & processing results', iter_count)
                task_gen = ((_remote_task, (n_iter, self.task_callable), {}) for n_iter in range(iter_start, iter_stop))
                for future in self.work_manager.submit_as_completed(task_gen, self.max_queue_len):
                    n_iter, result = future.get_result(discard=True)
                    if self.crawler is not None:
                        self.crawler.process_iter_result(n_iter, result)
                    pi.progress += 1
            finally:
                pi.new_operation('Finalizing')
                self.crawler.finalize()


def entry_point():
    WCrawl().main()


if __name__ == '__main__':
    entry_point()



"""
Routines for implementing Letteri et al.'s macrostate-to-macrostate rate calculations
using extrapolation to steady-state populations from average rate matrices

Internally, "labeled" objects (bin populations labeled by history, rate matrix elements labeled
by history) are stored as nested arrays -- e.g. rates[initial_label, final_label, initial_bin, final_bin].
These are converted to the flat forms required for, say, eigenvalue calculations internally, and the
results converted back. This is because these conversions are not expensive, and saves users of
this code from having to know how the flattened indexing works (something I screwed up all too
easily during development) -- mcz
"""

import logging
import warnings

import numpy as np
import scipy.linalg

from westpa.core.data_manager import weight_dtype

from ._kinetics import (
    calculate_labeled_fluxes,  # @UnresolvedImport
    calculate_labeled_fluxes_alllags,  # @UnresolvedImport
    labeled_flux_to_rate,  # @UnresolvedImport
    nested_to_flat_matrix,
    nested_to_flat_vector,  # @UnresolvedImport
    flat_to_nested_vector,
    _reduce_labeled_rate_matrix_to_macro,
)  # @UnresolvedImport


log = logging.getLogger(__name__)


class ConsistencyWarning(UserWarning):
    pass


def get_steady_state(rates):
    '''Get steady state solution for a rate matrix. As an optimization, returns the
    flattened labeled population vector (of length nstates*nbins); to convert to the
    nested vector used for storage, use nested_to_flat_vector().'''

    rates = rates.copy()

    # Convert to a transition probability matrix
    for i in range(rates.shape[0]):
        rowsum = rates[i, :].sum()
        if rowsum > 0:
            rates[i, :] = rates[i, :] / rowsum
        else:
            if rates[:, i].sum() != 0:
                warnings.warn('sink microstate in rate matrix', ConsistencyWarning)
            rates[i, :] = 0

    try:
        vals, vecs = scipy.linalg.eig(rates.T)
    except Exception:
        log.debug('exception obtaining eigenvectors', exc_info=True)
        return None

    vals = np.abs(vals)
    log.debug('eigenvalues: {!r}'.format(list(reversed(sorted(vals)))))
    asort = np.argsort(vals, kind='stable')
    vec = vecs[:, asort[-1]]
    ss = np.abs(vec)

    ss /= ss.sum()
    return ss


def get_macrostate_rates(labeled_rates, labeled_pops, extrapolate=True):
    '''Using a labeled rate matrix and labeled bin populations, calculate the steady state
    probability distribution and consequent state-to-state rates.

    Returns ``(ss, macro_rates)``, where ``ss`` is the steady-state probability distribution
    and ``macro_rates`` is the state-to-state rate matrix.'''

    nstates, nbins = labeled_pops.shape

    rates = nested_to_flat_matrix(labeled_rates)

    # Find steady-state solution
    if extrapolate:
        ss = get_steady_state(rates)
        if ss is None:
            warnings.warn('no well-defined steady state; using average populations', ConsistencyWarning)
            ss = nested_to_flat_vector(labeled_pops)
    else:
        ss = nested_to_flat_vector(labeled_pops)

    macro_rates = _reduce_labeled_rate_matrix_to_macro(nstates, nbins, rates, ss)

    return flat_to_nested_vector(nstates, nbins, ss), macro_rates


def estimate_rates(
    nbins,
    state_labels,
    weights,
    parent_ids,
    bin_assignments,
    label_assignments,
    state_map,
    labeled_pops,
    all_lags=False,
    labeled_fluxes=None,
    labeled_rates=None,
    unlabeled_rates=None,
):
    '''Estimate fluxes and rates over multiple iterations. The number of iterations is determined by how many
    vectors of weights, parent IDs, bin assignments, and label assignments are passed.

    If ``all_lags`` is true, then the average is over all possible lags within the length-N window given, otherwise
    simply the length N lag.

    Returns labeled flux matrix, labeled rate matrix, and unlabeled rate matrix.'''

    assert len(weights) == len(parent_ids) == len(bin_assignments) == len(label_assignments)
    nstates = len(state_labels)
    nbins = labeled_pops.shape[1] - 1

    # Prepare output arrays
    if labeled_fluxes is None:
        labeled_fluxes = np.zeros((nstates, nstates, nbins, nbins), weight_dtype)
    else:
        labeled_fluxes.fill(0.0)

    if labeled_rates is None:
        labeled_rates = np.zeros_like(labeled_fluxes)
    else:
        labeled_rates.fill(0.0)

    if unlabeled_rates is None:
        unlabeled_rates = np.zeros((nbins, nbins), weight_dtype)
    else:
        unlabeled_rates.fill(0.0)

    # Loop over all possible windows to accumulate flux matrix
    # flux matrix is [initial_label][final_label][initial_bin][final_bin]
    if all_lags:
        twindow = calculate_labeled_fluxes_alllags(nstates, weights, parent_ids, bin_assignments, label_assignments, labeled_fluxes)
    else:
        twindow = calculate_labeled_fluxes(nstates, weights, parent_ids, bin_assignments, label_assignments, labeled_fluxes)
    labeled_fluxes /= twindow

    # Calculate rate matrix for this window, using populations from the last iteration (which correspond
    # to the weights that contribute to the flux matrix)
    labeled_flux_to_rate(labeled_fluxes, labeled_pops, labeled_rates)

    # Calculate an unlabeled rate matrix
    unlabeled_fluxes = np.sum(labeled_fluxes, axis=(0, 1))
    unlabeled_pops = labeled_pops.sum(axis=0)
    unlabeled_rates[...] = labeled_flux_to_rate(unlabeled_fluxes[None, None, :, :], unlabeled_pops[None, :])[0, 0]

    return labeled_fluxes, labeled_rates, unlabeled_rates



"""WEST Analyis framework -- an unholy mess of classes exploiting each other"""

from . import atool
from .atool import WESTAnalysisTool
from .base_mixin import ArgumentError, AnalysisMixin
from .binning import BinningMixin
from .data_reader import WESTDataReaderMixin, ExtDataReaderMixin, BFDataManager
from .iter_range import IterRangeMixin
from .kinetics import KineticsAnalysisMixin
from .mcbs import MCBSMixin
from .output import CommonOutputMixin
from .plotting import PlottingMixin
from .trajwalker import TrajWalker
from .transitions import TransitionAnalysisMixin, TransitionEventAccumulator, BFTransitionAnalysisMixin


__all__ = [
    'atool',
    'AnalysisMixin',
    'ArgumentError',
    'WESTAnalysisTool',
    'IterRangeMixin',
    'WESTDataReaderMixin',
    'ExtDataReaderMixin',
    'BFDataManager',
    'BinningMixin',
    'MCBSMixin',
    'TrajWalker',
    'TransitionAnalysisMixin',
    'TransitionEventAccumulator',
    'BFTransitionAnalysisMixin',
    'KineticsAnalysisMixin',
    'CommonOutputMixin',
    'PlottingMixin',
]



import argparse
import logging
import traceback

import westpa
import westpa.work_managers as work_managers
from westpa.work_managers import make_work_manager


log = logging.getLogger('w_run')


def entry_point():
    parser = argparse.ArgumentParser('w_run', 'start/continue a WEST simulation')
    westpa.rc.add_args(parser)
    parser.add_argument(
        '--oneseg',
        dest='only_one_segment',
        action='store_true',
        help='only propagate one segment (useful for debugging propagators)',
    )

    work_managers.environment.add_wm_args(parser)

    args = parser.parse_args()
    westpa.rc.process_args(args)
    work_managers.environment.process_wm_args(args)

    run_simulation()


def run_simulation():
    work_manager = westpa.rc.work_manager = make_work_manager()

    # Load the sim manager and other drivers
    sim_manager = westpa.rc.get_sim_manager()
    system = westpa.rc.get_system_driver()
    data_manager = westpa.rc.get_data_manager()
    we_driver = westpa.rc.get_we_driver()
    propagator = westpa.rc.get_propagator()

    propagator.system = system
    data_manager.system = system
    we_driver.system = system

    sim_manager.data_manager = data_manager
    sim_manager.system = system
    sim_manager.propagator = propagator
    sim_manager.we_driver = we_driver

    with work_manager:
        if work_manager.is_master:
            work_manager.install_sigint_handler()
            sim_manager.load_plugins()

            log.debug('preparing run')
            sim_manager.prepare_run()

            try:
                log.debug('beginning run')
                sim_manager.run()

                log.debug('finalizing run')
                sim_manager.finalize_run()
            except KeyboardInterrupt:
                westpa.rc.pstatus('interrupted; shutting down')
            except Exception as e:
                westpa.rc.pstatus('exception caught; shutting down')
                if str(e) != '':
                    log.error(f'error message: {e}')
                log.error(traceback.format_exc())
        else:
            work_manager.run()


if __name__ == '__main__':
    entry_point()



import logging
import multiprocessing
import queue
import sys
import threading

from .core import WorkManager, WMFuture
import westpa.work_managers as work_managers

log = logging.getLogger(__name__)


class Task:
    def __init__(self, fn, args, kwargs, future):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.future = future

    def run(self):
        try:
            result = self.fn(*self.args, **self.kwargs)
        except Exception as e:
            self.future._set_exception(e, sys.exc_info()[2])
        else:
            self.future._set_result(result)


ShutdownSentinel = object()


class ThreadsWorkManager(WorkManager):
    '''A work manager using threads.'''

    @classmethod
    def from_environ(cls, wmenv=None):
        if wmenv is None:
            wmenv = work_managers.environment.default_env
        return cls(wmenv.get_val('n_workers', multiprocessing.cpu_count(), int))

    def __init__(self, n_workers=None):
        super().__init__()
        self.n_workers = n_workers or multiprocessing.cpu_count()
        self.workers = []
        self.task_queue = queue.Queue()

    def runtask(self, task_queue):
        while True:
            task = task_queue.get()
            if task is ShutdownSentinel:
                return
            else:
                task.run()

    def submit(self, fn, args=None, kwargs=None):
        ft = WMFuture()
        task = Task(fn, args if args is not None else (), kwargs if kwargs is not None else {}, ft)
        self.task_queue.put(task)
        return ft

    def startup(self):
        if not self.running:
            self.running = True
            self.workers = [
                threading.Thread(target=self.runtask, args=[self.task_queue], name='worker-{:d}'.format(i))
                for i in range(0, self.n_workers)
            ]
            for thread in self.workers:
                log.debug('starting thread {!r}'.format(thread))
                thread.start()

    def shutdown(self):
        if self.running:
            # Put one sentinel on the queue per worker, then wait for threads to terminate
            for i in range(0, self.n_workers):
                self.task_queue.put(ShutdownSentinel)
            for thread in self.workers:
                thread.join()
            self.running = False



import collections
import pytest

import numpy as np
from numpy.random import Generator, MT19937

from westpa.core.kinetics._kinetics import calc_rates, StreamingStats2D, StreamingStats1D
from westpa.core.kinetics.rate_averaging import tuple2stats


@pytest.fixture(autouse=True, scope='class')
def rng_setup(request):
    '''
    This function is run everytime a new class is initiated during a test in this module.
    '''
    request.cls.rng = Generator(MT19937())


class TestRateAverating:
    def test_tuple2stats2D(self):
        StreamingStatsTuple = collections.namedtuple('StreamingStatsTuple', ['M1', 'M2', 'n'])
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins, nbins))
        mask = np.zeros((nbins, nbins), np.uint8)

        rate_stats = StreamingStats2D((nbins, nbins))
        for d in data:
            rate_stats.update(d, mask)

        c_rate_stats = StreamingStatsTuple(rate_stats.M1, rate_stats.M2, rate_stats.n)

        rate_stats2 = tuple2stats(c_rate_stats)

        assert np.allclose(rate_stats.M1, rate_stats2.M1)
        assert np.allclose(rate_stats.M2, rate_stats2.M2)
        assert np.allclose(rate_stats.n, rate_stats2.n)
        assert np.allclose(rate_stats.mean, rate_stats2.mean)
        assert np.allclose(rate_stats.var, rate_stats2.var)

    def test_tuple2stats1D(self):
        StreamingStatsTuple = collections.namedtuple('StreamingStatsTuple', ['M1', 'M2', 'n'])
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins))
        mask = np.zeros((nbins,), np.uint8)

        rate_stats = StreamingStats1D(nbins)
        for d in data:
            rate_stats.update(d, mask)

        c_rate_stats = StreamingStatsTuple(rate_stats.M1, rate_stats.M2, rate_stats.n)

        rate_stats2 = tuple2stats(c_rate_stats)

        assert np.allclose(rate_stats.M1, rate_stats2.M1)
        assert np.allclose(rate_stats.M2, rate_stats2.M2)
        assert np.allclose(rate_stats.n, rate_stats2.n)
        assert np.allclose(rate_stats.mean, rate_stats2.mean)
        assert np.allclose(rate_stats.var, rate_stats2.var)


class TestKinetics:
    def test_calc_rates(self):
        nbins = 100
        mask = np.zeros((nbins, nbins), np.uint8)

        flux_matrix = self.rng.normal(size=(nbins, nbins))
        rate_matrix = np.zeros_like(flux_matrix)

        population_vector = self.rng.random(size=(nbins,)) + 0.0001
        population_vector[[0, 2, 5]] = 0.0

        calc_rates(flux_matrix, population_vector, rate_matrix, mask)

        expected_mask = np.zeros((nbins, nbins), np.uint8)
        expected_mask[[0, 2, 5], :] = 1

        nzi = np.where(population_vector != 0.0)[0]
        expected_rate = np.zeros_like(flux_matrix)
        expected_rate[nzi, :] = flux_matrix[nzi, :] / population_vector[nzi][:, np.newaxis]

        assert np.all(mask == expected_mask)
        assert np.allclose(rate_matrix, expected_rate)

        # row_sum = rate_matrix.sum(axis=1)
        # for k in xrange(nbins):
        # print(row_sum[k])
        # assert np.allclose(row_sum[k], 0.0) or np.allclose(row_sum[k], 1.0)


class TestStreamingStats2D:
    def test_nomask(self):
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins, nbins))
        mask = np.zeros((nbins, nbins), np.uint8)

        rate_stats = StreamingStats2D((nbins, nbins))
        for d in data:
            rate_stats.update(d, mask)

        assert np.allclose(rate_stats.mean, data.mean(axis=0))
        assert np.allclose(
            rate_stats.var,
            data.var(
                axis=0,
            ),
        )

    def test_nomask_groups(self):
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins, nbins))
        mask = np.zeros((nbins, nbins), np.uint8)

        rate_stats1 = StreamingStats2D((nbins, nbins))
        for d in data[: (nbins // 2)]:
            rate_stats1.update(d, mask)

        rate_stats2 = StreamingStats2D((nbins, nbins))
        for d in data[(nbins // 2) : nbins]:
            rate_stats2.update(d, mask)

        rate_stats3 = rate_stats1 + rate_stats2
        rate_stats1 += rate_stats2

        assert np.allclose(rate_stats1.mean, data.mean(axis=0))
        assert np.allclose(rate_stats1.var, data.var(axis=0))

        assert np.allclose(rate_stats3.mean, data.mean(axis=0))
        assert np.allclose(rate_stats3.var, data.var(axis=0))

    def test_with_mask(self):
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins, nbins))
        mask = self.rng.integers(2, size=data.shape, dtype=np.uint8)

        rate_stats = StreamingStats2D((nbins, nbins))
        for di, d in enumerate(data):
            rate_stats.update(d, mask[di])

        data_masked = np.ma.array(data, mask=mask)

        assert np.allclose(rate_stats.mean, data_masked.mean(axis=0).filled(fill_value=0.0))
        assert np.allclose(rate_stats.var, data_masked.var(axis=0).filled(fill_value=0.0))

    def test_with_mask_groups(self):
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins, nbins))
        mask = self.rng.integers(2, size=data.shape, dtype=np.uint8)

        rate_stats1 = StreamingStats2D((nbins, nbins))
        for di, d in enumerate(data[: (nbins // 2)]):
            rate_stats1.update(d, mask[di])

        rate_stats2 = StreamingStats2D((nbins, nbins))
        for di, d in enumerate(data[(nbins // 2) :]):
            rate_stats2.update(d, mask[di])

        rate_stats3 = rate_stats1 + rate_stats2
        rate_stats1 += rate_stats2

        data_masked = np.ma.array(data, mask=mask)

        assert np.allclose(rate_stats1.mean, data_masked.mean(axis=0).filled(fill_value=0.0))
        assert np.allclose(rate_stats1.var, data_masked.var(axis=0).filled(fill_value=0.0))

        assert np.allclose(rate_stats3.mean, data_masked.mean(axis=0).filled(fill_value=0.0))
        assert np.allclose(rate_stats3.var, data_masked.var(axis=0).filled(fill_value=0.0))


class TestStreamingStats1D:
    def test_nomask(self):
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins))
        mask = np.zeros((nbins,), np.uint8)

        rate_stats = StreamingStats1D(nbins)
        for d in data:
            rate_stats.update(d, mask)

        assert np.allclose(rate_stats.mean, data.mean(axis=0))
        assert np.allclose(rate_stats.var, data.var(axis=0))

    def test_nomask_groups(self):
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins))
        mask = np.zeros((nbins,), np.uint8)

        rate_stats1 = StreamingStats1D(nbins)
        for d in data[: (nbins // 2)]:
            rate_stats1.update(d, mask)

        rate_stats2 = StreamingStats1D(nbins)
        for d in data[(nbins // 2) :]:
            rate_stats2.update(d, mask)

        rate_stats3 = rate_stats1 + rate_stats2
        rate_stats1 += rate_stats2

        assert np.allclose(rate_stats1.mean, data.mean(axis=0))
        assert np.allclose(rate_stats1.var, data.var(axis=0))

        assert np.allclose(rate_stats3.mean, data.mean(axis=0))
        assert np.allclose(rate_stats3.var, data.var(axis=0))

    def test_with_mask(self):
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins))
        mask = self.rng.integers(2, size=data.shape, dtype=np.uint8)

        rate_stats = StreamingStats1D(nbins)
        for di, d in enumerate(data):
            rate_stats.update(d, mask[di])

        data_masked = np.ma.array(data, mask=mask)

        assert np.allclose(rate_stats.mean, data_masked.mean(axis=0).filled(fill_value=0.0))
        assert np.allclose(rate_stats.var, data_masked.var(axis=0).filled(fill_value=0.0))

    def test_with_mask_groups(self):
        nbins = 100
        nsets = 10
        data = self.rng.normal(size=(nsets, nbins))
        mask = self.rng.integers(2, size=data.shape, dtype=np.uint8)

        rate_stats1 = StreamingStats1D(nbins)
        for di, d in enumerate(data[: (nbins // 2)]):
            rate_stats1.update(d, mask[di])

        rate_stats2 = StreamingStats1D(nbins)
        for di, d in enumerate(data[(nbins // 2) :]):
            rate_stats2.update(d, mask[di])

        rate_stats3 = rate_stats1 + rate_stats2
        rate_stats1 += rate_stats2

        data_masked = np.ma.array(data, mask=mask)

        assert np.allclose(rate_stats1.mean, data_masked.mean(axis=0).filled(fill_value=0.0))
        assert np.allclose(rate_stats1.var, data_masked.var(axis=0).filled(fill_value=0.0))

        assert np.allclose(rate_stats3.mean, data_masked.mean(axis=0).filled(fill_value=0.0))
        assert np.allclose(rate_stats3.var, data_masked.var(axis=0).filled(fill_value=0.0))



import numpy as np
import pickle
import westpa

from westpa.core.propagators.executable import npy_data_loader, pickle_data_loader, aux_data_loader, ExecutablePropagator
from westpa.core.segment import Segment


class Test_Executable:
    '''Class to test the propagator executable.'''

    def test_data_config(self, ref_executable):
        '''Test if the config is initialized correctly, where the executable propagator dataset options are set with the data manager options.'''

        # Make the rc and executable read the config file.
        westpa.rc.read_config(filename='west_implicit.cfg')
        executable = ExecutablePropagator(rc=westpa.rc)

        assert 'displacement' in executable.data_info
        assert executable.data_info['displacement']['loader'] == npy_data_loader

    def test_legacy_data_config(self, ref_executable):
        '''Test if the dataset config is initialized correctly using the legacy part, where propagator datasets have to be specified twice.'''

        # Make the rc and executable read the config file.
        westpa.rc.read_config(filename='west.cfg')
        executable = ExecutablePropagator(rc=westpa.rc)

        assert 'displacement' in executable.data_info
        assert executable.data_info['displacement']['loader'] == aux_data_loader


class Test_Loaders:
    '''Class to test if npy_data_loader and pickle_date_loader are able to successfully add data into a dummy segment object.'''

    def test_npy_loader(self, ref_idtype):
        '''Test if data loaded with npy_data_loader is consistent.'''

        test_segment = Segment()

        npy_data_loader('test', self.correct_pkl, test_segment, False)

        with open(self.correct_pkl, 'rb') as f:
            ref_array = pickle.load(f)

        test_array = test_segment.data['test'][:]

        assert np.array_equal(test_array, ref_array)

    def test_pickle_loader(self, ref_idtype):
        '''Test if data loaded with npy_data_loader is consistent.'''

        test_segment = Segment()

        pickle_data_loader('test', self.correct_pkl, test_segment, False)

        with open(self.correct_pkl, 'rb') as f:
            ref_array = pickle.load(f)

        test_array = test_segment.data['test'][:]

        assert np.array_equal(test_array, ref_array)




# A cythoned version of the original function of the stats_process function,
# based on _kinetics.pyx

from __future__ import print_function,division
import cython
import numpy
import numpy as np
import h5py
from scipy.sparse import csgraph
import warnings
from collections import Counter
cimport numpy
cimport numpy as np
cimport scipy.linalg
cimport scipy.linalg.cython_lapack as cl
import scipy.linalg
from libc.math cimport isnan

ctypedef numpy.uint16_t index_t
ctypedef numpy.float64_t weight_t
ctypedef numpy.uint8_t bool_t
ctypedef numpy.int64_t trans_t
ctypedef numpy.uintp_t uint_t # 32 bits on 32-bit systems, 64 bits on 64-bit systems
ctypedef unsigned short Ushort
ctypedef double complex Cdouble

weight_dtype = numpy.float64
index_dtype = numpy.uint16
bool_dtype = numpy.bool_
intc_dtype = numpy.intc

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef stats_process(numpy.ndarray[index_t, ndim=2] bin_assignments,
                    numpy.ndarray[weight_t, ndim=1] weights,
                    numpy.ndarray[weight_t, ndim=2] fluxes,
                    numpy.ndarray[weight_t, ndim=1] populations,
                    numpy.ndarray[trans_t, ndim=2] trans,
                    numpy.ndarray[index_t, ndim=2] mask,
                    str interval='timepoint'                        ):
    cdef:
        Py_ssize_t i,k
        index_t ibin,fbin,nsegs,npts
    nsegs = bin_assignments.shape[0]
    npts = bin_assignments.shape[1]

    if interval == 'timepoint':
        for i in xrange(0,npts - 1):
            for k in xrange(nsegs):
                ibin = bin_assignments[k,i]
                fbin = bin_assignments[k, i + 1]

                if mask[k, 0] == 1:
                    continue

                w = weights[k]

                fluxes[ibin, fbin] += w
                trans[ibin, fbin] += 1
                populations[ibin] += w
        return

    if interval == 'iteration':
        for k in xrange(nsegs):
            # Should this be 0?
            # .... this should super be 0.  What?
            ibin = bin_assignments[k,0]
            fbin = bin_assignments[k, npts - 1]

            if mask[k, 0] == 1:
                continue

            w = weights[k]

            fluxes[ibin, fbin] += w
            trans[ibin, fbin] += 1
            populations[ibin] += w
        return

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int normalize(weight_t[:,:] m, Py_ssize_t nfbins) nogil:

    cdef:
        weight_t row_sum
        Py_ssize_t x, y

    for y in range(nfbins):
        row_sum = 0
        for x in range(nfbins):
            row_sum += m[y,x]
        if row_sum != 0:
            for x in range(nfbins):
                m[y,x] /= row_sum
    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef reweight_for_c(rows, cols, obs, flux, insert, indices, nstates, nbins, state_labels, state_map, nfbins, istate, jstate, stride, bin_last_state_map, bin_state_map, return_obs, obs_threshold=1):



    # Instead of pulling in start and stop, we'll pull in a list of indices.
    # This way, it should support the bootstrap.
    cdef:
        int[:] _rows, _cols, _obs, _ins, _nrows, _ncols, _nobs, _visited
        long[:]  _bin_last_state_map
        weight_t[:] _flux, _total_pop, _rw_bin_probs, _nflux, _rw_color_probs, _rw_state_probs
        double[:] _eigvals, _eigvalsi
        int n_trans, _nstates, lind, _nfbins, _stride, _obs_threshold, nnz, nlind, i, j, _istate, _jstate
        Ushort[:] _indices, _bin_state_map
        weight_t _return_value

        Ushort[:] _new_indices

        weight_t[:,:] _total_fluxes, _transition_matrix, _rw_state_flux, _strong_transition_matrix
        double[:,:] _WORK, _eigvecs
        int[:,:] _total_obs, _graph
        #bint _return_flux, _return_states, _return_color
        str _return_obs
        #double[:] eigvals, eigvalsi
        #double[:,:] eigvecs, WORK



    # CREATE NUMPY ARRAYS
    # This is a temporary measure that fixes some segfaults, which implies I'm probably off by
    # a little bit.  Memory heavy, but whatever.
    # It breaks depending on things, so I need to root that out.  Clearly, nnz is larger than that.
    _flux = flux
    nnz = len(flux)
    lind = indices.shape[0]
    nlind = indices.shape[0]*stride
    total_fluxes = np.zeros((nfbins, nfbins), weight_dtype)
    total_obs = np.zeros((nfbins, nfbins), intc_dtype)
    transition_matrix = np.zeros((nfbins, nfbins), weight_dtype)
    strong_transition_matrix = np.zeros((nfbins, nfbins), weight_dtype)
    visited = np.zeros((nfbins), intc_dtype)
    graph = np.zeros((nfbins, nfbins+1), dtype=intc_dtype)
    rw_bin_probs = np.zeros(nfbins, weight_dtype)
    new_indices = np.zeros(((nlind)), dtype=indices.dtype)
    rw_state_flux = np.zeros((nstates, nstates), np.float64)
    state_flux = np.zeros((nstates, nstates), weight_dtype)
    eigvals = np.zeros((nfbins), np.float64)
    eigvalsi = np.zeros((nfbins), np.float64)
    eigvecs = np.zeros((nfbins, nfbins), np.float64)
    WORK = np.zeros((nfbins*4, nfbins*4), np.float64)
    rw_color_probs = np.zeros((nstates), weight_dtype)
    rw_state_probs = np.zeros((nbins), weight_dtype)


    # CREATE MEMORYVIEWS
    # These are for what we sent in...
    _rows = rows
    _cols = cols
    _obs = obs
    _flux = flux
    _ins = insert
    _nstates = nstates
    # ... these are for functions we'll be using.
    _total_fluxes = total_fluxes
    _total_obs = total_obs
    _transition_matrix = transition_matrix
    _strong_transition_matrix = strong_transition_matrix
    _nfbins = nfbins
    _stride = stride
    _obs_threshold = obs_threshold
    _indices = indices
    _new_indices = new_indices
    _rw_state_flux = rw_state_flux
    _rw_bin_probs = rw_bin_probs
    _eigvals = eigvals
    _eigvalsi = eigvalsi
    _eigvals = eigvals
    _eigvecs = eigvecs
    _WORK = WORK
    _graph = graph
    _visited = visited
    _rw_color_probs = rw_color_probs
    _rw_state_probs = rw_state_probs
    _bin_last_state_map = bin_last_state_map
    _bin_state_map = bin_state_map
    _istate = istate
    _jstate = jstate
    _return_obs = return_obs


    #NOGIL
    # Reconstruct dataset.  We're just passing the same thing back and forth between functions.
    with nogil:
        for i in range(_nfbins):
            for j in range(1, _nfbins+1):
                _graph[i, j] = _nfbins
        regenerate_subsampled_indices(_indices, _new_indices, lind, _stride)
        accumulate_fluxes(_rows, _cols, _obs, _flux, _ins, _new_indices, nnz, _transition_matrix, nlind)
        accumulate_obs(_rows, _cols, _obs, _flux, _ins, _new_indices, nnz, _total_obs, nlind)

        remove_under_obs(_transition_matrix, _total_obs, _obs_threshold, _nfbins)
        normalize(_transition_matrix, _nfbins)
        steadystate_solve(_transition_matrix, _strong_transition_matrix, _rw_bin_probs, _nfbins, _eigvals, _eigvalsi, _eigvecs, _WORK, _graph, _visited)

        for i in range(_nfbins):
            _rw_color_probs[_bin_last_state_map[i]] += _rw_bin_probs[i]
            _rw_state_probs[_bin_state_map[i]] += _rw_bin_probs[i]


        calc_state_flux(_transition_matrix, _rw_bin_probs, _bin_last_state_map, _bin_state_map, _nstates, _rw_state_flux, _nfbins)

    # This allows us to use the same function for all three types.
    # Return conditional fluxes.
    if _return_obs == b'F':
        _return_value = _rw_state_flux[_istate,_jstate]
        if isnan(_return_value) is True:
            return 0.0
        else:
            return _return_value
    # Return state probabilities.
    elif _return_obs == b'S':
        _return_value = _rw_state_probs[_istate]
        if isnan(_return_value) is True:
            return 0.0
        else:
            return _return_value
    # Return color (ensemble) probabilities
    elif _return_obs == b'C':
        _return_value = _rw_color_probs[_istate]
        if isnan(_return_value) is True:
            return 0.0
        else:
            return _return_value
    # Return the rates.
    elif _return_obs == b'R':
        if _rw_color_probs[_istate] != 0.0:
            _return_value = (_rw_state_flux[_istate,_jstate] / (_rw_color_probs[_istate] / (_rw_color_probs[_istate] + _rw_color_probs[_jstate])))
            if isnan(_return_value) is True:
                return 0.0
            else:
                return _return_value
        else:
            # We have no ensemble probability, and as such, cannot have a flux.
            return 0.0
    # Return the populations.
    elif _return_obs == b'P':
        return rw_bin_probs

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int regenerate_subsampled_indices(Ushort[:] iin, Ushort[:] iout, int ilen, int stride) nogil:

    cdef:
        int i, si

    # go over the range of all indices within iin
    for i in range(ilen):
        # Run over the length of the stride.
        for si in range(stride):
            iout[(i*stride)+si] = iin[i] + si

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int accumulate_fluxes(int[:] hrows, int[:] hcols, int[:] hobs, weight_t[:] hflux, int[:] hins, Ushort[:] iterations, Py_ssize_t nnz, weight_t[:,:] total_fluxes, int itermax) nogil:

    cdef:
        index_t curriter, elem, iiter, ipop
        long ilem

    curriter = 0

    for iter in range(itermax):
        iiter = iterations[iter]
        for ilem in range(hins[iiter], hins[iiter+1]):
            # Not sure if this is necessary, here...
            if ilem < nnz and iiter+1 < itermax:
                total_fluxes[hrows[ilem], hcols[ilem]] += hflux[ilem]

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int accumulate_obs(int[:] hrows, int[:] hcols, int[:] hobs, weight_t[:] hflux, int[:] hins, Ushort[:] iterations, Py_ssize_t nnz, int[:,:] total_obs, int itermax) nogil:

    cdef:
        index_t curriter, elem, iiter, ipop
        long ilem

    curriter = 0

    for iter in range(itermax):
        iiter = iterations[iter]
        for ilem in range(hins[iiter], hins[iiter+1]):
            if ilem < nnz and iiter+1 < itermax:
                total_obs[hrows[ilem], hcols[ilem]] += hobs[ilem]

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int remove_under_obs(weight_t[:,:] flux, int[:,:] obs, int threshold, int nbins) nogil:

    cdef:
        int x, y

    for x in range(nbins):
        for y in range(nbins):
            if obs[x,y] < threshold:
                flux[x,y] = 0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int calc_state_flux(weight_t[:, :] trans_matrix, weight_t[:] bin_probs, long[:] bin_last_state_map, Ushort[:] bin_state_map, int nstates, weight_t[:,:] state_flux, int K_shape) nogil:

    cdef:
        int i, j, ii, jj


    for i in range(K_shape):
        for j in range(K_shape):

            ii = bin_last_state_map[i]
            jj = bin_state_map[j]

            if jj != nstates:
                state_flux[ii, jj] += (trans_matrix[i, j] * bin_probs[i])

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int steadystate_solve(weight_t[:,:] K, weight_t[:,:] K_mod, weight_t[:] bin_prob, int K_shape, double[:] eigvals, double[:] eigvalsi, double[:,:] eigvecs, double[:,:] WORK, int[:,:] graph, int[:] visited) nogil:

    cdef:
        int[:] components
        int[:,:] _graph
        double max, eigsum
        int n_components, components_assignments, largest_component, maxi, x, y, n, INFO, LWORK, i, j

        # POINTERS

        int  *_INFO, *_K_shape, *_LWORK
        double *_K_mod, *_eigvals, *_eigvecs, *_WORK, *_eigvalsi
    _K_shape = &K_shape
    _INFO = &INFO
    _LWORK = &LWORK
    INFO = 0
    LWORK = K_shape * 4
    _K_mod = &K_mod[0,0]
    _eigvals = &eigvals[0]
    _eigvecs = &eigvecs[0,0]
    _eigvalsi = &eigvalsi[0]
    _WORK = &WORK[0,0]
    _graph = graph

    for i in range(K_shape):
        if visited[i] == 0:
            visited[i] = 1
            return_strong_component(K, K_shape, _graph, i, i, visited)
    n = 0
    for i in range(K_shape):
        if graph[i, 0] >= graph[n, 0]:
            n = i
    # I suspect this may be giving us issues?
    #components = _graph[n, :K_shape+1]

    maxi = 0
    eigsum = 0.0
    # This all works!
    for x in range(K_shape):
        #i = components[x+1]
        i = graph[n, x+1]
        for y in range(K_shape):
            #j = components[y+1]
            j = graph[n, y+1]
            if i != K_shape and j != K_shape:
                K_mod[i, j] = K[i, j]
    normalize(K_mod, K_shape)
    cl.dgeev('N', 'V', _K_shape, _K_mod, _K_shape, _eigvals, _eigvalsi, _eigvecs, _K_shape, _eigvecs, _K_shape, _WORK, _LWORK, _INFO)
    for x in range(K_shape):
        if x == 0:
            max = eigvals[0]
            maxi = x
        else:
            if max < eigvals[x]:
                max = eigvals[x]
                maxi = x
    # We need to go over the whole range and pick out non K_shape elements.
    # This probably no longer needs to be done, now...
    for i in range(K_shape):
        #x = components[i+1]
        x = graph[n, i+1]
        if x != K_shape:
            #eigsum += eigvecs[maxi, components[i+1]]
            eigsum += eigvecs[maxi, x]
    for i in range(K_shape):
        #x = components[i+1]
        x = graph[n, i+1]
        if x != K_shape:
            #bin_prob[components[i+1]] = eigvecs[maxi, components[i+1]]
            #bin_prob[components[i+1]] /= eigsum
            bin_prob[x] = eigvecs[maxi, x]
            bin_prob[x] /= eigsum

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int return_strong_component(weight_t[:,:] K, int K_shape, int[:, :] graph, int i, int z, int[:] visited) nogil:

    cdef:
        int j, y

    if graph[z, 0] == 0:
        graph[z, 0] += 1
        graph[z, 1] = i
    for j in xrange(K_shape):
        if i != j:
            if K[i, j] > 0.0:
                # Strongly connected!
                if visited[j] == 0:
                    graph[z, 0] += 1
                    y = graph[z, 0]
                    graph[z, y] = j
                    # We only want to call it when we haven't visited it before.
                    # We don't want to call, THEN modify and check.  Otherwise, we could be doing many calls.
                    visited[j] = 1
                    return_strong_component(K, K_shape, graph, j, z, visited)


    return 0



import math

import numpy as np


class Segment:
    '''A class wrapping segment data that must be passed through the work manager or data manager.
    Most fields are self-explanatory.  One item worth noting is that a negative parent ID means that
    the segment starts from the initial state with ID -(segment.parent_id+1)
    '''

    SEG_STATUS_UNSET = 0
    SEG_STATUS_PREPARED = 1
    SEG_STATUS_COMPLETE = 2
    SEG_STATUS_FAILED = 3

    SEG_INITPOINT_UNSET = 0
    SEG_INITPOINT_CONTINUES = 1
    SEG_INITPOINT_NEWTRAJ = 2

    SEG_ENDPOINT_UNSET = 0
    SEG_ENDPOINT_CONTINUES = 1
    SEG_ENDPOINT_MERGED = 2
    SEG_ENDPOINT_RECYCLED = 3

    statuses = {}
    initpoint_types = {}
    endpoint_types = {}

    status_names = {}
    initpoint_type_names = {}
    endpoint_type_names = {}

    # convenience functions for binning
    @staticmethod
    def initial_pcoord(segment):
        'Return the initial progress coordinate point of this segment.'
        return segment.pcoord[0]

    @staticmethod
    def final_pcoord(segment):
        'Return the final progress coordinate point of this segment.'
        return segment.pcoord[-1]

    def __init__(
        self,
        n_iter=None,
        seg_id=None,
        weight=None,
        endpoint_type=None,
        parent_id=None,
        wtg_parent_ids=None,
        pcoord=None,
        status=None,
        walltime=None,
        cputime=None,
        data=None,
    ):
        # NaNs appear sometimes if a WEST program is terminated unexpectedly; replace with zero
        walltime = 0.0 if walltime is None or math.isnan(walltime) else walltime
        cputime = 0.0 if cputime is None or math.isnan(cputime) else cputime

        # the int() and float() calls are required so that new-style string formatting doesn't barf
        # assuming that the respective fields are actually strings, probably after implicitly
        # calling __str__() on them.  Not sure if this is a numpy, h5py, or python problem
        self.n_iter = int(n_iter) if n_iter is not None else None
        self.seg_id = int(seg_id) if seg_id is not None else None
        self.status = int(status) if status is not None else None
        self.parent_id = int(parent_id) if parent_id is not None else None
        self.endpoint_type = int(endpoint_type) if endpoint_type else self.SEG_ENDPOINT_UNSET

        self.weight = float(weight) if weight is not None else None
        self.wtg_parent_ids = set(wtg_parent_ids or ())

        self.pcoord = np.asarray(pcoord) if pcoord is not None else None
        self.walltime = walltime
        self.cputime = cputime
        self.data = data if data else {}

    def __repr__(self):
        return '<%s(%s) n_iter=%r seg_id=%r weight=%r parent_id=%r wtg_parent_ids=%r pcoord[0]=%r pcoord[-1]=%r>' % (
            self.__class__.__name__,
            hex(id(self)),
            self.n_iter,
            self.seg_id,
            self.weight,
            self.parent_id,
            tuple(self.wtg_parent_ids or ()),
            self.pcoord[0] if self.pcoord is not None else None,
            self.pcoord[-1] if self.pcoord is not None else None,
        )

    @property
    def initpoint_type(self):
        if self.parent_id < 0:
            return Segment.SEG_INITPOINT_NEWTRAJ
        else:
            return Segment.SEG_INITPOINT_CONTINUES

    @property
    def initial_state_id(self):
        if self.parent_id < 0:
            return -(self.parent_id + 1)
        else:
            return None

    status_text = property((lambda s: s.status_names[s.status]))
    endpoint_type_text = property((lambda s: s.endpoint_type_names[s.endpoint_type]))


Segment.statuses.update({_attr: getattr(Segment, _attr) for _attr in dir(Segment) if _attr.startswith('SEG_STATUS_')})
Segment.initpoint_types.update({_attr: getattr(Segment, _attr) for _attr in dir(Segment) if _attr.startswith('SEG_INITPOINT_')})
Segment.endpoint_types.update({_attr: getattr(Segment, _attr) for _attr in dir(Segment) if _attr.startswith('SEG_ENDPOINT_')})

Segment.status_names.update({getattr(Segment, _attr): _attr for _attr in dir(Segment) if _attr.startswith('SEG_STATUS_')})
Segment.initpoint_type_names.update(
    {getattr(Segment, _attr): _attr for _attr in dir(Segment) if _attr.startswith('SEG_INITPOINT_')}
)
Segment.endpoint_type_names.update({getattr(Segment, _attr): _attr for _attr in dir(Segment) if _attr.startswith('SEG_ENDPOINT_')})



import logging

from westpa.oldtools.aframe import AnalysisMixin

log = logging.getLogger(__name__)


class CommonOutputMixin(AnalysisMixin):
    def __init__(self):
        super().__init__()

        include_args = self.include_args.setdefault('CommonOutputMixin', {})
        include_args.setdefault('suppress_headers', True)
        include_args.setdefault('print_bin_labels', True)

        self.output_suppress_headers = False
        self.output_print_bin_labels = False

    def add_common_output_args(self, parser_or_group):
        if self.include_args['CommonOutputMixin']['suppress_headers']:
            parser_or_group.add_argument(
                '--noheaders',
                dest='suppress_headers',
                action='store_true',
                help='Do not include headers in text output files (default: include headers)',
            )
        if self.include_args['CommonOutputMixin']['print_bin_labels']:
            parser_or_group.add_argument(
                '--binlabels',
                dest='print_bin_labels',
                action='store_true',
                help='Print bin labels in output files, if available (default: do not print bin labels)',
            )

    def process_common_output_args(self, args):
        if self.include_args['CommonOutputMixin']['suppress_headers']:
            self.output_suppress_headers = bool(args.suppress_headers)
        if self.include_args['CommonOutputMixin']['print_bin_labels']:
            self.output_print_bin_labels = bool(args.print_bin_labels)



import argparse
import logging

import numpy as np

import westpa
from westpa.core.segment import Segment
from westpa.core.states import InitialState
from westpa.core.data_manager import n_iter_dtype, seg_id_dtype

log = logging.getLogger('w_fork')


def entry_point():
    parser = argparse.ArgumentParser(
        'w_fork',
        description='''\
    Prepare a new weighted ensemble simulation from an existing one at a particular
    point. A new HDF5 file is generated. In the case of executable propagation,
    it is the user's responsibility to prepare the new simulation directory
    appropriately, particularly making the old simulation's restart data from the
    appropriate iteration available as the new simulations initial state data; a
    mapping of old simulation segment to new simulation initial states is
    created, both in the new HDF5 file and as a flat text file, to aid in this.
    Target states and basis states for the new simulation are taken from those
    in the original simulation.
    ''',
    )

    westpa.rc.add_args(parser)
    parser.add_argument(
        '-i',
        '--input',
        dest='input_h5file',
        help='''Create simulation from the given INPUT_H5FILE (default: read from
                                configuration file.''',
    )
    parser.add_argument(
        '-I',
        '--iteration',
        dest='n_iter',
        type=int,
        help='''Take initial distribution for new simulation from iteration N_ITER
                                (default: last complete iteration).''',
    )
    parser.add_argument(
        '-o',
        '--output',
        dest='output_h5file',
        default='forked.h5',
        help='''Save new simulation HDF5 file as OUTPUT (default: %(default)s).''',
    )
    parser.add_argument(
        '--istate-map',
        default='istate_map.txt',
        help='''Write text file describing mapping of existing segments to new initial
                                states in ISTATE_MAP (default: %(default)s).''',
    )
    parser.add_argument('--no-headers', action='store_true', help='''Do not write header to ISTATE_MAP''')
    args = parser.parse_args()
    westpa.rc.process_args(args)

    # Open old HDF5 file
    dm_old = westpa.rc.new_data_manager()
    if args.input_h5file:
        dm_old.we_h5filename = args.input_h5file
    dm_old.open_backing(mode='r')

    # Get iteration if necessary
    n_iter = args.n_iter or dm_old.current_iteration - 1

    # Create and open new HDF5 file
    dm_new = westpa.rc.new_data_manager()
    dm_new.we_h5filename = args.output_h5file
    dm_new.prepare_backing()
    dm_new.open_backing()

    # Copy target states
    target_states = dm_old.get_target_states(n_iter)
    dm_new.save_target_states(target_states, n_iter)

    # Copy basis states
    basis_states = dm_old.get_basis_states(n_iter)
    dm_new.create_ibstate_group(basis_states, n_iter=1)

    # Transform old segments into initial states and new segments
    # We produce one initial state and one corresponding
    # new segment for each old segment. Further adjustment
    # can be accomplished by using w_binning.
    old_iter_group = dm_old.get_iter_group(n_iter)
    old_index = old_iter_group['seg_index'][...]
    old_pcoord_ds = old_iter_group['pcoord']
    n_segments = old_pcoord_ds.shape[0]
    pcoord_len = old_pcoord_ds.shape[1]
    pcoord_ndim = old_pcoord_ds.shape[2]
    old_final_pcoords = old_pcoord_ds[:, pcoord_len - 1, :]

    istates = dm_new.create_initial_states(n_segments, n_iter=1)
    segments = []
    state_map_dtype = np.dtype([('old_n_iter', n_iter_dtype), ('old_seg_id', seg_id_dtype), ('new_istate_id', seg_id_dtype)])
    state_map = np.empty((n_segments,), dtype=state_map_dtype)
    state_map['old_n_iter'] = n_iter

    for iseg, (index_row, pcoord) in enumerate(zip(old_index, old_final_pcoords)):
        istate = istates[iseg]
        istate.iter_created = 0
        istate.iter_used = 1
        istate.istate_type = InitialState.ISTATE_TYPE_RESTART
        istate.istate_status = InitialState.ISTATE_STATUS_PREPARED
        istate.pcoord = pcoord

        segment = Segment(
            n_iter=1,
            seg_id=iseg,
            weight=index_row['weight'],
            parent_id=-(istate.state_id + 1),
            wtg_parent_ids=[-(istate.state_id + 1)],
            status=Segment.SEG_STATUS_PREPARED,
        )
        segment.pcoord = np.zeros((pcoord_len, pcoord_ndim), dtype=pcoord.dtype)
        segment.pcoord[0] = pcoord
        segments.append(segment)
        state_map[iseg]['old_seg_id'] = iseg
        state_map[iseg]['new_istate_id'] = istate.state_id

    dm_new.update_initial_states(istates, n_iter=0)
    dm_new.prepare_iteration(n_iter=1, segments=segments)

    # Update current iteration and close both files
    dm_new.current_iteration = 1
    dm_new.close_backing()
    dm_old.close_backing()

    # Write state map
    istate_map_file = open(args.istate_map, 'wt')
    if not args.no_headers:
        istate_map_file.write('# mapping from previous segment IDs to new initial states\n')
        istate_map_file.write('# generated by w_fork\n')
        istate_map_file.write('# column 0: old simulation n_iter\n')
        istate_map_file.write('# column 1: old simulation seg_id\n')
        istate_map_file.write('# column 2: new simulation initial state ID\n')

    for row in state_map:
        istate_map_file.write(
            '{old_n_iter:20d}    {old_seg_id:20d}    {new_istate_id:20d}\n'.format(
                old_n_iter=int(row['old_n_iter']), old_seg_id=int(row['old_seg_id']), new_istate_id=int(row['new_istate_id'])
            )
        )


if __name__ == '__main__':
    entry_point()



import os
import pytest

import h5py
import numpy as np
from scipy.spatial.distance import cdist

from westpa.core.binning.assign import (
    RectilinearBinMapper,
    PiecewiseBinMapper,
    FuncBinMapper,
    VectorizingFuncBinMapper,
    VoronoiBinMapper,
    RecursiveBinMapper,
)
from westpa.core.binning.assign import coord_dtype
from westpa.core.binning.mab import MABBinMapper, map_mab


REFERENCE_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'refs')


class TestRectilinearBinMapper:
    def test1dAssign(self):
        bounds = [0.0, 1.0, 2.0, 3.0]
        coords = np.array([0, 0.5, 1.5, 1.6, 2.0, 2.0, 2.9])[:, None]

        assigner = RectilinearBinMapper([bounds])
        assert (assigner.assign(coords) == [0, 0, 1, 1, 2, 2, 2]).all()

    def test2dAssign(self):
        boundaries = [(-1, -0.5, 0, 0.5, 1), (-1, -0.5, 0, 0.5, 1)]
        coords = np.array([(-0.75, -0.75), (-0.25, -0.25), (0, 0), (0.25, 0.25), (0.75, 0.75), (-0.25, 0.75), (0.25, -0.75)])
        assigner = RectilinearBinMapper(boundaries)

        """bin structure: [(a,b), (c,d)] => x in [a,b), y in [c, d)
        0:[(-1, -0.5), (-1, -0.5)]
        1:[(-1, -0.5), (-0.5, 0)]
        2:[(-1, -0.5), (0, 0.5)]
        3:[(-1, -0.5), (0.5, 1)]
        4:[(-0.5, 0), (-1, -0.5)]
        5:[(-0.5, 0), (-0.5, 0)]
        6:[(-0.5, 0), (0, 0.5)]
        7:[(-0.5, 0), (0.5, 1)]
        8:[(0, 0.5), (-1, -0.5)]
        9:[(0, 0.5), (-0.5, 0)]
        10:[(0, 0.5), (0, 0.5)]
        11:[(0, 0.5), (0.5, 1)]
        12:[(0.5, 1), (-1, -0.5)]
        13:[(0.5, 1), (-0.5, 0)]
        14:[(0.5, 1), (0, 0.5)]
        15:[(0.5, 1), (0.5, 1)]"""

        assert (assigner.assign(coords) == [0, 5, 10, 10, 15, 7, 8]).all()


class TestPiecewiseBinMapper:
    def test_bin_mapping(self):
        coords = np.array([[-0.5], [0.0], [0.5]], dtype=np.float32)
        fr1 = lambda x: x < 0
        fr2 = lambda x: x >= 0
        pm = PiecewiseBinMapper([fr1, fr2])
        assert list(pm.assign(coords)) == [0, 1, 1]


class TestFuncBinMapper:
    @staticmethod
    def fn(coords, mask, output):
        output[mask & (coords[:, 0] <= 0.5)] = 0
        output[mask & (coords[:, 0] > 0.5)] = 1

    def test_fmapper(self):
        mapper = FuncBinMapper(self.fn, 2)
        coords = np.array([0.0, 0.1, 0.5, 0.7])
        coords.shape = (coords.shape[0], 1)
        print(repr(coords))
        output = mapper.assign(coords)
        print(repr(output))
        assert list(output) == [0, 0, 0, 1]


class TestVectorizingFuncBinMapper:
    @staticmethod
    def fn(coord):
        if coord[0] <= 0.5:
            return 0
        else:
            return 1

    def test_vmapper(self):
        mapper = VectorizingFuncBinMapper(self.fn, 2)
        coords = np.array([0.0, 0.1, 0.5, 0.7])
        coords.shape = (coords.shape[0], 1)
        output = mapper.assign(coords)
        assert list(output) == [0, 0, 0, 1]


class TestVoronoiBinMapper:
    @staticmethod
    def distfunc(coordvec, centers):
        if coordvec.ndim < 2:
            new_coordvec = np.empty((1, coordvec.shape[0]), dtype=coord_dtype)
            new_coordvec[0, :] = coordvec[:]
            coordvec = new_coordvec
        distmat = np.require(cdist(coordvec, centers), dtype=coord_dtype)
        return distmat[0, :]

    def test_vmapper(self):
        centers = np.array([[0, 0], [2, 2]], dtype=coord_dtype)
        coords = np.array([[0, 0], [2, 2], [0.9, 0.9], [1.1, 1.1]], dtype=coord_dtype)

        mapper = VoronoiBinMapper(self.distfunc, centers)
        output = mapper.assign(coords)
        assert list(output) == [0, 1, 0, 1]


class TestNestingBinMapper:
    # pass
    '''
    0                            1                      2
    +----------------------------+----------------------+
    |            0.5             |         1.5          |
    | +-----------+------------+ | +--------+---------+ |
    | |    0.25   |            | | |        |         | |
    | | +---+---+ |            | | |        |         | |
    | | |   |   | |            | | |        |         | |
    | | +---+---+ |            | | |        |         | |
    | +-----------+------------+ | +--------+---------+ |
    +---------------------------------------------------+
    '''

    @staticmethod
    def fn1(coords, mask, output):
        test = coords[:, 0] < 1
        output[mask & test] = 0
        output[mask & ~test] = 1

    @staticmethod
    def fn2(coords, mask, output):
        test = coords[:, 0] < 0.5
        output[mask & test] = 0
        output[mask & ~test] = 1

    @staticmethod
    def fn3(coords, mask, output):
        test = coords[:, 0] < 0.25
        output[mask & test] = 0
        output[mask & ~test] = 1

    @staticmethod
    def fn4(coords, mask, output):
        test = coords[:, 0] < 1.5
        output[mask & test] = 0
        output[mask & ~test] = 1

    def testOuterMapper(self):
        '''
        0                            1                      2
        +----------------------------+----------------------+
        |              0             |          1           |
        +---------------------------------------------------+
        '''

        mapper = FuncBinMapper(self.fn1, 2)
        rmapper = RecursiveBinMapper(mapper)
        coords = np.array([[0.1], [0.2], [0.3], [0.4], [0.6], [1.1]])
        output = rmapper.assign(coords)
        assert list(output) == [0, 0, 0, 0, 0, 1]

    def testOuterMapperWithOffset(self):
        '''
        0                            1                      2
        +----------------------------+----------------------+
        |              1             |          2           |
        +---------------------------------------------------+
        '''

        mapper = FuncBinMapper(self.fn1, 2)
        rmapper = RecursiveBinMapper(mapper, 1)
        coords = np.array([[0.1], [0.2], [0.3], [0.4], [0.6], [1.1]])
        output = rmapper.assign(coords)
        assert list(output) == [1, 1, 1, 1, 1, 2]

    def testSingleRecursion(self):
        '''
        0                            1                      2
        +----------------------------+----------------------+
        |            0.5             |                      |
        | +-----------+------------+ |                      |
        | |           |            | |                      |
        | |     1     |     2      | |          0           |
        | |           |            | |                      |
        | |           |            | |                      |
        | +-----------+------------+ |                      |
        +---------------------------------------------------+
        '''

        outer_mapper = FuncBinMapper(self.fn1, 2)
        inner_mapper = FuncBinMapper(self.fn2, 2)
        rmapper = RecursiveBinMapper(outer_mapper)
        rmapper.add_mapper(inner_mapper, [0.5])
        assert rmapper.nbins == 3
        coords = np.array([[0.1], [0.2], [0.3], [0.4], [0.6], [1.1]])
        output = rmapper.assign(coords)
        assert list(output) == [1, 1, 1, 1, 2, 0]

    def testDeepRecursion(self):
        '''
        0                            1                      2
        +----------------------------+----------------------+
        |            0.5             |                      |
        | +-----------+------------+ |                      |
        | |    0.25   |            | |                      |
        | | +---+---+ |            | |           0          |
        | | | 2 | 3 | |     1      | |                      |
        | | +---+---+ |            | |                      |
        | +-----------+------------+ |                      |
        +---------------------------------------------------+
        '''

        outer_mapper = FuncBinMapper(self.fn1, 2)
        middle_mapper = FuncBinMapper(self.fn2, 2)
        inner_mapper = FuncBinMapper(self.fn3, 2)

        rmapper = RecursiveBinMapper(outer_mapper)
        rmapper.add_mapper(middle_mapper, [0.5])
        rmapper.add_mapper(inner_mapper, [0.25])

        assert rmapper.nbins == 4
        coords = np.array([[0.1], [0.2], [0.3], [0.4], [0.6], [1.1]])
        output = rmapper.assign(coords)
        assert list(output) == [2, 2, 3, 3, 1, 0]

    def testSideBySideRecursion(self):
        '''
        0                            1                      2
        +----------------------------+----------------------+
        |            0.5             |         1.5          |
        | +-----------+------------+ | +--------+---------+ |
        | |           |            | | |        |         | |
        | |     0     |     1      | | |   2    |    3    | |
        | |           |            | | |        |         | |
        | |           |            | | |        |         | |
        | +-----------+------------+ | +--------+---------+ |
        +---------------------------------------------------+
        '''
        outer_mapper = FuncBinMapper(self.fn1, 2)
        middle_mapper1 = FuncBinMapper(self.fn2, 2)
        middle_mapper2 = FuncBinMapper(self.fn4, 2)

        rmapper = RecursiveBinMapper(outer_mapper)
        rmapper.add_mapper(middle_mapper1, [0.5])
        rmapper.add_mapper(middle_mapper2, [1.5])

        assert rmapper.nbins == 4
        coords = np.array([[0.1], [0.2], [0.3], [0.4], [0.6], [1.1], [1.6]])
        output = rmapper.assign(coords)
        print('OUTPUT', output)
        assert list(output) == [0, 0, 0, 0, 1, 2, 3]

    def testMegaComplexRecursion(self):
        '''
        0                            1                      2
        +----------------------------+----------------------+
        |            0.5             |         1.5          |
        | +-----------+------------+ | +--------+---------+ |
        | |    0.25   |            | | |        |         | |
        | | +---+---+ |            | | |        |         | |
        | | | 1 | 2 | |     0      | | |   3    |    4    | |
        | | +---+---+ |            | | |        |         | |
        | +-----------+------------+ | +--------+---------+ |
        +---------------------------------------------------+
        '''
        outer_mapper = FuncBinMapper(self.fn1, 2)
        middle_mapper1 = FuncBinMapper(self.fn2, 2)
        middle_mapper2 = FuncBinMapper(self.fn4, 2)
        inner_mapper = FuncBinMapper(self.fn3, 2)

        rmapper = RecursiveBinMapper(outer_mapper)
        rmapper.add_mapper(middle_mapper1, [0.5])
        rmapper.add_mapper(inner_mapper, [0.25])
        rmapper.add_mapper(middle_mapper2, [1.5])

        assert rmapper.nbins == 5
        coords = np.array([[0.1], [0.2], [0.3], [0.4], [0.6], [1.1], [1.6]])
        output = rmapper.assign(coords)
        assert list(output) == [1, 1, 2, 2, 0, 3, 4]

    # TODO: Fix this test
    @pytest.mark.xfail(reason="known error in assign")
    def test2dRectilinearRecursion(self):
        '''
         0                            1                      2
         +----------------------------+----------------------+
         |                            |         1.5          |
         |                            | +--------+---------+ |
         |                            | |        |         | |
         |             0              | |   3    |   4     | |
         |                            | |        |         | |
         |                            | |        |         | |
         |                            | +--------+---------+ |
        1+---------------------------------------------------+
         |            0.5             |                      |
         | +-----------+------------+ |                      |
         | |           |            | |                      |
         | |    1      |     2      | |           5          |
         | |           |            | |                      |
         | |           |            | |                      |
         | +-----------+------------+ |                      |
        2+---------------------------------------------------+

        '''

        outer_mapper = RectilinearBinMapper([[0, 1, 2], [0, 1, 2]])

        upper_right_mapper = RectilinearBinMapper([[1, 1.5, 2], [0, 1]])
        lower_left_mapper = RectilinearBinMapper([[0, 0.5, 1], [1, 2]])

        rmapper = RecursiveBinMapper(outer_mapper)
        rmapper.add_mapper(upper_right_mapper, [1.5, 0.5])
        rmapper.add_mapper(lower_left_mapper, [0.5, 1.5])

        pairs = [(0.5, 0.5), (1.25, 0.5), (1.75, 0.5), (0.25, 1.5), (0.75, 1.5), (1.5, 1.5)]

        assert rmapper.nbins == 6
        assignments = rmapper.assign(pairs)
        expected = [0, 3, 4, 1, 2, 5]
        print('PAIRS', pairs)
        print('LABELS', list(rmapper.labels))
        print('EXPECTED', expected)
        print('OUTPUT  ', assignments)
        assert (assignments == expected).all()


# Following section is for MAB Testing
@pytest.fixture(scope='class')
def gen_input_mab_data_fixture(request):
    '''A fixture to assign `gen_input_mab_data`'s return as a class attribute for tests.'''
    request.cls.input_mab_data = gen_input_mab_data()


def gen_input_mab_data():
    '''Function to generate test data for MABBinMapper Testing.'''
    # Create synthetic test data: a 2D grid of points with Gaussian weights distribution
    N_point_2d = 15
    n_dim_2d = 2
    N_total_2d = N_point_2d**n_dim_2d
    coords_2d = np.zeros((N_total_2d, 2))
    for i in range(N_total_2d):
        coords_2d[i, 0] = i % N_point_2d
        coords_2d[i, 1] = i // N_point_2d
    coords_2d /= N_point_2d

    # generate weights as a 2D gaussian where the max is at the center of the 2D space
    weights_2d = np.zeros(N_total_2d)
    for i in range(N_total_2d):
        weights_2d[i] = np.exp(-((coords_2d[i, 0] - 1 / 2) ** 2 + (coords_2d[i, 1] - 1 / 2) ** 2) / (1 / 2) ** 2)
    weights_2d /= np.sum(weights_2d)

    # Tile coords
    allcoords_2d_grid = np.ones((N_total_2d, 4))
    allcoords_2d_grid[:, :2] = coords_2d
    allcoords_2d_grid[:, 2] = weights_2d
    allcoords_2d_grid = np.tile(allcoords_2d_grid, (2, 1))
    allcoords_2d_grid[0:N_total_2d, 3] = 0

    # Create deterministic 3D array of grid points, again with 3D Gaussian distribution of weights
    N_point_3d = 15
    n_dim_3d = 3
    N_total_3d = N_point_3d**n_dim_3d
    coords_3d = np.zeros((N_total_3d, 3))
    for i in range(N_total_3d):
        coords_3d[i, 0] = i % N_point_3d
        coords_3d[i, 1] = (i // N_point_3d) % N_point_3d
        coords_3d[i, 2] = i // (N_point_3d**2)
    coords_3d /= N_point_3d

    # Generate weights as a n_dim gaussian where the max is at the center of the 3D space
    weights_3d = np.zeros(N_total_3d)
    for i in range(N_total_3d):
        weights_3d[i] = np.exp(
            -((coords_3d[i, 0] - 1 / 2) ** 2 + (coords_3d[i, 1] - 1 / 2) ** 2 + (coords_3d[i, 2] - 1 / 2) ** 2) / (1 / 2) ** 2
        )
    weights_3d /= np.sum(weights_3d)

    # Tile coords
    allcoords_3d_grid = np.ones((N_total_3d, n_dim_3d + 2))
    allcoords_3d_grid[:, :n_dim_3d] = coords_3d
    allcoords_3d_grid[:, n_dim_3d] = weights_3d
    allcoords_3d_grid = np.tile(allcoords_3d_grid, (2, 1))
    allcoords_3d_grid[0:N_total_3d, n_dim_3d + 1] = 0

    # Lastly, test bin assignment with 2D coordinates and weights on pseudorandom deterministic Gaussian points
    N_total_random = 300
    rng = np.random.default_rng(seed=0)
    coords_random = rng.normal(loc=[0.5, 0.5], scale=0.25, size=(N_total_random, n_dim_2d))

    # Generate weights as a n_dim sine curve with given wavelength plus some deterministic noise
    weights_random = np.zeros(N_total_random)
    wavelength = 0.25
    noise_level = 0.1
    for i in range(n_dim_2d):
        weights_random += np.sin(2 * np.pi * coords_random[:, i] / wavelength) + noise_level * np.cos(
            4 * 2 * np.pi * coords_random[:, i] / wavelength
        )
    weights_random = np.abs(weights_random) / np.sum(np.abs(weights_random))

    # Tile coords
    allcoords_2d_gauss = np.ones((N_total_random, 4))
    allcoords_2d_gauss[:, :2] = coords_random
    allcoords_2d_gauss[:, 2] = weights_random
    allcoords_2d_gauss = np.tile(allcoords_2d_gauss, (2, 1))
    allcoords_2d_gauss[0:N_total_random, 3] = 0

    return {
        'allcoords_2d_grid': allcoords_2d_grid,
        'allcoords_3d_grid': allcoords_3d_grid,
        'allcoords_2d_gauss': allcoords_2d_gauss,
    }


@pytest.fixture(scope='class')
def ref_mab_results(request):
    '''Function for reading the reference datasets from `refs/mab_assignments_ref.h5`.'''
    # Setting up empty things dictionary with corresponding keys
    request.cls.ref_mab_results = {}
    test_keys = ['2d_grid', '3d_grid', '2d_gauss']

    # Loading in the three sets of reference data from the reference file into a dictionary
    with h5py.File(os.path.join(REFERENCE_PATH, 'mab_assignments_ref.h5'), 'r') as f:
        for key in test_keys:
            request.cls.ref_mab_results[key] = [f[f'{key}/test_result_{i:d}'][...] for i in range(len(f[key]))]


class TestMABBinMapper:
    '''Test class for MABBinMapper'''

    @pytest.mark.parametrize(
        'nbins, direction, skip, bottleneck, ref_value',
        [
            ([5, 1], [1, 86], [0, 0], True, 9),
            ([5, 5], [0, 0], [0, 0], True, 33),
            ([5, 5, 5], [0, 0, 0], [0, 0, 0], True, 137),
            ([5, 5], [0, 0], [1, 0], True, 9),
            ([5, 1], [1, 86], [0, 0], False, 6),
        ],
        ids=['5x1, direction=[1,86]', '5x5', '5x5x5', '5x5, skip=[1,0]', '5x1, direction=[1,86], no bottleneck'],
    )
    def test_determine_total_bins(self, nbins, direction, skip, bottleneck, ref_value):
        '''Runs through different configurations to see if mab reports the correct number of total bins'''
        mab = MABBinMapper([5])  # Creating a dummy MABBinMapper

        assert mab.determine_total_bins(nbins_per_dim=nbins, direction=direction, skip=skip, bottleneck=bottleneck) == ref_value

    @pytest.mark.parametrize(
        "nbins_per_dim, direction, bottleneck, skip, ref_index",
        [
            ([2, 2], [1, 1], False, [0, 0], 0),
            ([2, 2], [0, 0], False, [0, 0], 1),
            ([2, 2], [-1, -1], True, [0, 0], 2),
            ([2, 2], [0, 0], True, [0, 0], 3),
            ([2, 2], [0, 0], True, [0, 1], 4),
            ([2, 2], [0, 0], True, [1, 1], 5),
            ([2, 2], [86, 0], True, [0, 0], 6),
            ([2, 2], [86, 86], False, [0, 0], 7),
        ],
        ids=[
            'direction=[1,1], no bottleneck',
            'direction=[0,0], no bottleneck',
            'direction=[-1,-1]',
            'direction=[0,0]',
            'direction=[0,0], skip=[0,1]',
            'direction=[0,0], skip=[1,1]',
            'direction=[86,0]',
            'direction=[86,86], no bottleneck',
        ],
    )
    def test_2x2_2d_grid_mab_bin_assignments(
        self, gen_input_mab_data_fixture, ref_mab_results, nbins_per_dim, direction, bottleneck, skip, ref_index
    ):
        '''Test MABBinMapper with 2x2 linear section on 2D space'''
        allcoords = self.input_mab_data['allcoords_2d_grid']
        N_total = allcoords.shape[0] // 2
        mask = np.full((N_total * 2), True)
        output = np.zeros((N_total * 2), dtype=int)
        output = map_mab(
            coords=allcoords,
            mask=mask,
            output=output,
            nbins_per_dim=nbins_per_dim,
            direction=direction,
            bottleneck=bottleneck,
            skip=skip,
        )
        assert np.all(output[:N_total] == output[N_total:]), "Expected first half of bin assignments to equal second half"
        assert np.all(
            output == self.ref_mab_results['2d_grid'][ref_index]
        ), f"Unexpected 2D grid MAB bin assignments with direction={direction}, bottleneck={bottleneck}, and skip={skip}"

    @pytest.mark.parametrize(
        "nbins_per_dim, direction, bottleneck, skip, ref_index",
        [
            ([2, 2, 2], [0, 0, 0], False, [0, 0, 0], 0),
            ([2, 2, 2], [0, 0, 0], True, [0, 0, 0], 1),
        ],
        ids=[
            'no bottleneck',
            'with bottleneck',
        ],
    )
    def test_2x2x2_3d_grid_mab_bin_assignments(
        self, gen_input_mab_data_fixture, ref_mab_results, nbins_per_dim, direction, bottleneck, skip, ref_index
    ):
        '''Test MABBinMapper with 2x2x2 linear section on 3D space'''
        allcoords = self.input_mab_data['allcoords_3d_grid']
        N_total = allcoords.shape[0] // 2
        mask = np.full((N_total * 2), True)
        output = list(np.zeros((N_total * 2), dtype=int))
        output = map_mab(
            coords=allcoords,
            mask=mask,
            output=output,
            nbins_per_dim=nbins_per_dim,
            direction=direction,
            bottleneck=bottleneck,
            skip=skip,
        )
        assert output[:N_total] == output[N_total:], "Expected first half of bin assignments to equal second half"
        assert output == list(
            self.ref_mab_results['3d_grid'][ref_index]
        ), f"Unexpected 3D grid MAB bin assignments with direction={direction}, bottleneck={bottleneck}, and skip={skip}"

    @pytest.mark.parametrize(
        "nbins_per_dim, direction, bottleneck, skip, ref_index",
        [
            ([2, 2], [0, 0], True, [0, 0], 0),
            ([2, 2], [0, 0], True, [0, 1], 1),
            ([2, 2], [86, -1], True, [0, 0], 2),
        ],
        ids=[
            'direction=[0,0], no skip',
            'direction=[0,0], skip=[0,1]',
            'direction=[86,-1], no skip',
        ],
    )
    def test_2x2_2d_gaussian_mab_bin_assignments(
        self, gen_input_mab_data_fixture, ref_mab_results, nbins_per_dim, direction, bottleneck, skip, ref_index
    ):
        '''Test MABBinMapper with 2x2 linear section on a 2D Gaussian space'''
        allcoords = self.input_mab_data['allcoords_2d_gauss']
        N_total = allcoords.shape[0] // 2
        mask = np.full((N_total * 2), True)
        output = np.zeros((N_total * 2), dtype=int)
        output = map_mab(
            coords=allcoords,
            mask=mask,
            output=output,
            nbins_per_dim=nbins_per_dim,
            direction=direction,
            bottleneck=bottleneck,
            skip=skip,
        )
        assert np.all(output[:N_total] == output[N_total:]), "Expected first half of bin assignments to equal second half"
        assert np.all(
            output == self.ref_mab_results['2d_gauss'][ref_index]
        ), f"Unexpected 2D Gaussian MAB bin assignments with direction={direction}, bottleneck={bottleneck}, and skip={skip}"


def output_mab_reference():
    '''
    Function to generate the reference test files for the MAB tests.

    To run this, run `python -c 'from test_binning import output_mab_reference; output_mab_reference()'`
    in the command line (assuming you're in the `tests/` folder).

    It will overwrite the file 'tests/refs/mab_assignments_ref.h5' and create 11 png files in `tests/` visualizing the binning.
    '''
    import matplotlib.pyplot as plt

    try:
        os.mkdir('mab_tests')
    except FileExistsError:
        pass

    input_data = gen_input_mab_data()
    with h5py.File(f'{REFERENCE_PATH}/mab_assignments_ref.h5', 'w') as f:
        # 2D Grid
        for i, (nbins_per_dim, direction, bottleneck, skip) in enumerate(
            [
                ([2, 2], [1, 1], False, [0, 0]),
                ([2, 2], [0, 0], False, [0, 0]),
                ([2, 2], [-1, -1], True, [0, 0]),
                ([2, 2], [0, 0], True, [0, 0]),
                ([2, 2], [0, 0], True, [0, 1]),
                ([2, 2], [0, 0], True, [1, 1]),
                ([2, 2], [86, 0], True, [0, 0]),
                ([2, 2], [86, 86], False, [0, 0]),
            ]
        ):
            allcoords = input_data['allcoords_2d_grid']
            N_total = allcoords.shape[0] // 2
            mask = np.full((N_total * 2), True)
            output = np.zeros((N_total * 2), dtype=int)
            output = map_mab(
                coords=allcoords,
                mask=mask,
                output=output,
                nbins_per_dim=nbins_per_dim,
                direction=direction,
                bottleneck=bottleneck,
                skip=skip,
            )

            f.create_dataset(f'2d_grid/test_result_{i:d}', data=output)

            # Create a cmap with the same number of colors as the number of bins
            cmap = plt.cm.get_cmap('tab20', int(np.max(output) + 1))

            # Plot the synthetic data in 2D using a scatter plot
            # Include a cbar to shown the bin assignments
            plt.scatter(
                allcoords[:N_total, 0],
                allcoords[:N_total, 1],
                s=allcoords[:N_total, 2] * 10000,
                c=output[:N_total],
                cmap=cmap,
                vmin=-0.5,
                vmax=int(np.max(output)) + 0.5,
            )
            plt.colorbar(label='Bin ID', ticks=range(int(np.max(output) + 1)))
            plt.title(f'nbins_per_dim={nbins_per_dim}, direction={direction},\n bottleneck={bottleneck}, skip={skip}')
            plt.savefig(f'mab_tests/2d_grid_ref_result_{i}.png')
            plt.clf()

        # 3D grid
        for i, (nbins_per_dim, direction, bottleneck, skip) in enumerate(
            [
                ([2, 2, 2], [0, 0, 0], False, [0, 0, 0]),
                ([2, 2, 2], [0, 0, 0], True, [0, 0, 0]),
            ]
        ):
            allcoords = input_data['allcoords_3d_grid']
            N_total = allcoords.shape[0] // 2
            mask = np.full((N_total * 2), True)
            output = np.zeros((N_total * 2), dtype=int)
            output = map_mab(
                coords=allcoords,
                mask=mask,
                output=output,
                nbins_per_dim=nbins_per_dim,
                direction=direction,
                bottleneck=bottleneck,
                skip=skip,
            )

            f.create_dataset(f'3d_grid/test_result_{i:d}', data=output)

        # 2D Gaussian
        for i, (nbins_per_dim, direction, bottleneck, skip) in enumerate(
            [
                ([2, 2], [0, 0], True, [0, 0]),
                ([2, 2], [0, 0], True, [0, 1]),
                ([2, 2], [86, -1], True, [0, 0]),
            ]
        ):
            allcoords = input_data['allcoords_2d_gauss']
            N_total = allcoords.shape[0] // 2
            mask = np.full((N_total * 2), True)
            output = np.zeros((N_total * 2), dtype=int)
            output = map_mab(
                coords=allcoords,
                mask=mask,
                output=output,
                nbins_per_dim=nbins_per_dim,
                direction=direction,
                bottleneck=bottleneck,
                skip=skip,
            )

            f.create_dataset(f'2d_gauss/test_result_{i:d}', data=output)

            # Create a cmap with the same number of colors as the number of bins
            cmap = plt.cm.get_cmap('tab20', int(np.max(output) + 1))

            # Plot the synthetic data in 2D using a scatter plot
            # Include a cbar to shown the bin assignments
            plt.scatter(
                allcoords[:N_total, 0],
                allcoords[:N_total, 1],
                s=allcoords[:N_total, 2] * 10000,
                c=output[:N_total],
                cmap=cmap,
                vmin=-0.5,
                vmax=int(np.max(output)) + 0.5,
            )
            plt.colorbar(label='Bin ID', ticks=range(int(np.max(output) + 1)))
            plt.title(f'nbins_per_dim={nbins_per_dim}, direction={direction},\n bottleneck={bottleneck}, skip={skip}')
            plt.savefig(f'mab_tests/2d_gauss_ref_result_{i}.png')
            plt.clf()
    print("Reference data generated and saved to file.")



import logging
import math
import os

import numpy as np
from numpy import index_exp

from westpa.core.data_manager import seg_id_dtype, weight_dtype
from westpa.core.binning import index_dtype, assign_and_label, accumulate_labeled_populations
from westpa.tools import WESTParallelTool, WESTDataReader, WESTDSSynthesizer, BinMappingComponent, ProgressIndicatorComponent
import westpa
from westpa.core import h5io
from westpa.core.h5io import WESTPAH5File
from westpa.core.extloader import get_object

log = logging.getLogger('w_assign')


# Changes to keep it alive...
def parse_pcoord_value(pc_str):
    namespace = {'math': math, 'numpy': np, 'np': np, 'inf': float('inf')}

    arr = np.array(eval(pc_str, namespace))
    if arr.ndim == 0:
        arr.shape = (1, 1)
    elif arr.ndim == 1:
        arr.shape = (1,) + arr.shape
    else:
        raise ValueError('too many dimensions')
    return arr


def _assign_label_pop(
    n_iter, lb, ub, mapper, nstates, state_map, last_labels, parent_id_dsspec, weight_dsspec, pcoord_dsspec, subsample
):
    nbins = len(state_map) - 1
    parent_ids = parent_id_dsspec.get_iter_data(n_iter, index_exp[lb:ub])
    weights = weight_dsspec.get_iter_data(n_iter, index_exp[lb:ub])
    pcoords = pcoord_dsspec.get_iter_data(n_iter, index_exp[lb:ub])

    assignments, trajlabels, statelabels = assign_and_label(
        lb, ub, parent_ids, mapper.assign, nstates, state_map, last_labels, pcoords, subsample
    )
    pops = np.zeros((nstates + 1, nbins + 1), weight_dtype)
    accumulate_labeled_populations(weights, assignments, trajlabels, pops)
    return (assignments, trajlabels, pops, lb, ub, statelabels)


class WAssign(WESTParallelTool):
    prog = 'w_assign'
    description = '''\
Assign walkers to bins, producing a file (by default named "assign.h5")
which can be used in subsequent analysis.

For consistency in subsequent analysis operations, the entire dataset
must be assigned, even if only a subset of the data will be used. This
ensures that analyses that rely on tracing trajectories always know the
originating bin of each trajectory.


-----------------------------------------------------------------------------
Source data
-----------------------------------------------------------------------------

Source data is provided either by a user-specified function
(--construct-dataset) or a list of "data set specifications" (--dsspecs).
If neither is provided, the progress coordinate dataset ''pcoord'' is used.

To use a custom function to extract or calculate data whose probability
distribution will be calculated, specify the function in standard Python
MODULE.FUNCTION syntax as the argument to --construct-dataset. This function
will be called as function(n_iter,iter_group), where n_iter is the iteration
whose data are being considered and iter_group is the corresponding group
in the main WEST HDF5 file (west.h5). The function must return data which can
be indexed as [segment][timepoint][dimension].

To use a list of data set specifications, specify --dsspecs and then list the
desired datasets one-by-one (space-separated in most shells). These data set
specifications are formatted as NAME[,file=FILENAME,slice=SLICE], which will
use the dataset called NAME in the HDF5 file FILENAME (defaulting to the main
WEST HDF5 file west.h5), and slice it with the Python slice expression SLICE
(as in [0:2] to select the first two elements of the first axis of the
dataset). The ``slice`` option is most useful for selecting one column (or
more) from a multi-column dataset, such as arises when using a progress
coordinate of multiple dimensions.


-----------------------------------------------------------------------------
Specifying macrostates
-----------------------------------------------------------------------------

Optionally, kinetic macrostates may be defined in terms of sets of bins.
Each trajectory will be labeled with the kinetic macrostate it was most
recently in at each timepoint, for use in subsequent kinetic analysis.
This is required for all kinetics analysis (w_kintrace and w_kinmat).

There are three ways to specify macrostates:

  1. States corresponding to single bins may be identified on the command
     line using the --states option, which takes multiple arguments, one for
     each state (separated by spaces in most shells). Each state is specified
     as a coordinate tuple, with an optional label prepended, as in
     ``bound:1.0`` or ``unbound:(2.5,2.5)``. Unlabeled states are named
     ``stateN``, where N is the (zero-based) position in the list of states
     supplied to --states.

  2. States corresponding to multiple bins may use a YAML input file specified
     with --states-from-file. This file defines a list of states, each with a
     name and a list of coordinate tuples; bins containing these coordinates
     will be mapped to the containing state. For instance, the following
     file::

        ---
        states:
          - label: unbound
            coords:
              - [9.0, 1.0]
              - [9.0, 2.0]
          - label: bound
            coords:
              - [0.1, 0.0]

     produces two macrostates: the first state is called "unbound" and
     consists of bins containing the (2-dimensional) progress coordinate
     values (9.0, 1.0) and (9.0, 2.0); the second state is called "bound"
     and consists of the single bin containing the point (0.1, 0.0).

  3. Arbitrary state definitions may be supplied by a user-defined function,
     specified as --states-from-function=MODULE.FUNCTION. This function is
     called with the bin mapper as an argument (``function(mapper)``) and must
     return a list of dictionaries, one per state. Each dictionary must contain
     a vector of coordinate tuples with key "coords"; the bins into which each
     of these tuples falls define the state. An optional name for the state
     (with key "label") may also be provided.


-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, by default "assign.h5") contains the following
attributes datasets:

  ``nbins`` attribute
    *(Integer)* Number of valid bins. Bin assignments range from 0 to
    *nbins*-1, inclusive.

  ``nstates`` attribute
    *(Integer)* Number of valid macrostates (may be zero if no such states are
    specified). Trajectory ensemble assignments range from 0 to *nstates*-1,
    inclusive, when states are defined.

  ``/assignments`` [iteration][segment][timepoint]
    *(Integer)* Per-segment and -timepoint assignments (bin indices).

  ``/npts`` [iteration]
    *(Integer)* Number of timepoints in each iteration.

  ``/nsegs`` [iteration]
    *(Integer)* Number of segments in each iteration.

  ``/labeled_populations`` [iterations][state][bin]
    *(Floating-point)* Per-iteration and -timepoint bin populations, labeled
    by most recently visited macrostate. The last state entry (*nstates-1*)
    corresponds to trajectories initiated outside of a defined macrostate.

  ``/bin_labels`` [bin]
    *(String)* Text labels of bins.

When macrostate assignments are given, the following additional datasets are
present:

  ``/trajlabels`` [iteration][segment][timepoint]
    *(Integer)* Per-segment and -timepoint trajectory labels, indicating the
    macrostate which each trajectory last visited.

  ``/state_labels`` [state]
    *(String)* Labels of states.

  ``/state_map`` [bin]
    *(Integer)* Mapping of bin index to the macrostate containing that bin.
    An entry will contain *nbins+1* if that bin does not fall into a
    macrostate.

Datasets indexed by state and bin contain one more entry than the number of
valid states or bins. For *N* bins, axes indexed by bin are of size *N+1*, and
entry *N* (0-based indexing) corresponds to a walker outside of the defined bin
space (which will cause most mappers to raise an error). More importantly, for
*M* states (including the case *M=0* where no states are specified), axes
indexed by state are of size *M+1* and entry *M* refers to trajectories
initiated in a region not corresponding to a defined macrostate.

Thus, ``labeled_populations[:,:,:].sum(axis=1)[:,:-1]`` gives overall per-bin
populations, for all defined bins and
``labeled_populations[:,:,:].sum(axis=2)[:,:-1]`` gives overall
per-trajectory-ensemble populations for all defined states.


-----------------------------------------------------------------------------
Parallelization
-----------------------------------------------------------------------------

This tool supports parallelized binning, including reading/calculating input
data.


-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def __init__(self):
        super().__init__()

        # Parallel processing by default (this is not actually necessary, but it is
        # informative!)
        self.wm_env.default_work_manager = self.wm_env.default_parallel_work_manager

        self.data_reader = WESTDataReader()
        self.dssynth = WESTDSSynthesizer(default_dsname='pcoord')
        self.binning = BinMappingComponent()
        self.progress = ProgressIndicatorComponent()
        self.output_file = None
        self.output_filename = None
        self.states = []
        self.subsample = False

    def add_args(self, parser):
        self.data_reader.add_args(parser)
        self.binning.add_args(parser)
        self.dssynth.add_args(parser)

        sgroup = parser.add_argument_group('macrostate definitions').add_mutually_exclusive_group()
        sgroup.add_argument(
            '--states',
            nargs='+',
            metavar='STATEDEF',
            help='''Single-bin kinetic macrostate, specified by a coordinate tuple (e.g. '1.0' or '[1.0,1.0]'),
                            optionally labeled (e.g. 'bound:[1.0,1.0]'). States corresponding to multiple bins
                            must be specified with --states-from-file.''',
        )
        sgroup.add_argument(
            '--states-from-file',
            metavar='STATEFILE',
            help='''Load kinetic macrostates from the YAML file STATEFILE. See description
                            above for the appropriate structure.''',
        )
        sgroup.add_argument(
            '--states-from-function',
            metavar='STATEFUNC',
            help='''Load kinetic macrostates from the function STATEFUNC, specified as
                            module_name.func_name. This function is called with the bin mapper as an argument,
                            and must return a list of dictionaries {'label': state_label, 'coords': 2d_array_like}
                            one for each macrostate; the 'coords' entry must contain enough rows to identify all bins
                            in the macrostate.''',
        )

        agroup = parser.add_argument_group('other options')
        agroup.add_argument(
            '-o', '--output', dest='output', default='assign.h5', help='''Store results in OUTPUT (default: %(default)s).'''
        )
        agroup.add_argument(
            '--subsample',
            dest='subsample',
            action='store_const',
            const=True,
            help='''Determines whether or not the data should be subsampled.
                             This is rather useful for analysing steady state simulations.''',
        )
        agroup.add_argument(
            '--config-from-file',
            dest='config_from_file',
            action='store_true',
            help='''Load bins/macrostates from a scheme specified in west.cfg.''',
        )
        agroup.add_argument('--scheme-name', dest='scheme', help='''Name of scheme specified in west.cfg.''')

    def process_args(self, args):
        self.progress.process_args(args)
        self.data_reader.process_args(args)
        # Necessary to open the file to get the current iteration
        # if we want to use the mapper in the file
        self.data_reader.open(mode='r+')
        self.n_iter = self.data_reader.current_iteration
        # If we decide to use this option for iteration selection:
        # getattr(args,'bins_from_h5file',None) or self.data_reader.current_iteration

        with self.data_reader:
            self.dssynth.h5filename = self.data_reader.we_h5filename
            self.dssynth.process_args(args)
            if args.config_from_file is False:
                self.binning.set_we_h5file_info(self.n_iter, self.data_reader)
                self.binning.process_args(args)

        self.output_filename = args.output

        if args.config_from_file:
            if not args.scheme:
                raise ValueError('A scheme must be specified.')
            else:
                self.load_config_from_west(args.scheme)
        elif args.states:
            self.parse_cmdline_states(args.states)
        elif args.states_from_file:
            self.load_state_file(args.states_from_file)
        elif args.states_from_function:
            self.load_states_from_function(get_object(args.states_from_function, path=['.']))

        if self.states and len(self.states) < 2:
            raise ValueError('zero, two, or more macrostates are required')

        # self.output_file = WESTPAH5File(args.output, 'w', creating_program=True)
        log.debug('state list: {!r}'.format(self.states))

        self.subsample = args.subsample if args.subsample is not None else False

    def parse_cmdline_states(self, state_strings):
        states = []
        for istring, state_string in enumerate(state_strings):
            try:
                (label, coord_str) = state_string.split(':')
            except ValueError:
                label = 'state{}'.format(istring)
                coord_str = state_string
            coord = parse_pcoord_value(coord_str)
            states.append({'label': label, 'coords': coord})
        self.states = states

    def load_config_from_west(self, scheme):
        try:
            config = westpa.rc.config['west']['analysis']
        except Exception:
            raise ValueError('There is no configuration file specified.')
        ystates = config['analysis_schemes'][scheme]['states']
        self.states_from_dict(ystates)
        try:
            self.subsample = config['subsample']
        except Exception:
            pass
        from westpa.core._rc import bins_from_yaml_dict

        self.binning.mapper = bins_from_yaml_dict(config['analysis_schemes'][scheme]['bins'][0])
        path = os.path.join(os.getcwd(), config['directory'], scheme)
        try:
            os.mkdir(config['directory'])
            os.mkdir(path)
        except Exception:
            pass

        self.output_filename = os.path.join(path, 'assign.h5')

    def load_state_file(self, state_filename):
        import yaml

        ydict = yaml.load(open(state_filename, 'rt'), Loader=yaml.Loader)
        ystates = ydict['states']
        self.states_from_dict(ystates)

    def states_from_dict(self, ystates):
        states = []
        for istate, ystate in enumerate(ystates):
            state = {}
            state['label'] = ystate.get('label', 'state{}'.format(istate))
            # coords can be:
            #  - a scalar, in which case it is one bin, 1-D
            #  - a single list, which is rejected as ambiguous
            #  - a list of lists, which is a list of coordinate tuples
            coords = np.array(ystate['coords'])
            if coords.ndim == 0:
                coords.shape = (1, 1)
            elif coords.ndim == 1:
                raise ValueError(
                    'list {!r} is ambiguous (list of 1-d coordinates, or single multi-d coordinate?)'.format(ystate['coords'])
                )
            elif coords.ndim > 2:
                raise ValueError('coordinates must be 2-D')
            state['coords'] = coords
            states.append(state)
        self.states = states

    def load_states_from_function(self, statefunc):
        states = statefunc(self.binning.mapper)
        for istate, state in enumerate(states):
            state.setdefault('label', 'state{}'.format(istate))
            try:
                state['coords'] = np.array(state['coords'])
            except KeyError:
                raise ValueError('state function {!r} returned a state {!r} without coordinates'.format(statefunc, state))
        self.states = states
        log.debug('loaded states: {!r}'.format(self.states))

    def assign_iteration(self, n_iter, nstates, nbins, state_map, last_labels):
        '''Method to encapsulate the segment slicing (into n_worker slices) and parallel job submission
        Submits job(s), waits on completion, splices them back together
        Returns: assignments, trajlabels, pops for this iteration'''

        futures = []

        iter_group = self.data_reader.get_iter_group(n_iter)
        nsegs, npts = iter_group['pcoord'].shape[:2]
        n_workers = self.work_manager.n_workers or 1
        assignments = np.empty((nsegs, npts), dtype=index_dtype)
        trajlabels = np.empty((nsegs, npts), dtype=index_dtype)
        statelabels = np.empty((nsegs, npts), dtype=index_dtype)
        pops = np.zeros((nstates + 1, nbins + 1), dtype=weight_dtype)

        # Submit jobs to work manager
        blocksize = nsegs // n_workers
        if nsegs % n_workers > 0:
            blocksize += 1

        def task_gen():
            if __debug__:
                checkset = set()
            for lb in range(0, nsegs, blocksize):
                ub = min(nsegs, lb + blocksize)
                if __debug__:
                    checkset.update(set(range(lb, ub)))
                args = ()
                kwargs = dict(
                    n_iter=n_iter,
                    lb=lb,
                    ub=ub,
                    mapper=self.binning.mapper,
                    nstates=nstates,
                    state_map=state_map,
                    last_labels=last_labels,
                    parent_id_dsspec=self.data_reader.parent_id_dsspec,
                    weight_dsspec=self.data_reader.weight_dsspec,
                    pcoord_dsspec=self.dssynth.dsspec,
                    subsample=self.subsample,
                )
                yield (_assign_label_pop, args, kwargs)

                # futures.append(self.work_manager.submit(_assign_label_pop,
                # kwargs=)
            if __debug__:
                assert checkset == set(range(nsegs)), 'segments missing: {}'.format(set(range(nsegs)) - checkset)

        # for future in self.work_manager.as_completed(futures):
        for future in self.work_manager.submit_as_completed(task_gen(), queue_size=self.max_queue_len):
            assign_slice, traj_slice, slice_pops, lb, ub, state_slice = future.get_result(discard=True)
            assignments[lb:ub, :] = assign_slice
            trajlabels[lb:ub, :] = traj_slice
            statelabels[lb:ub, :] = state_slice
            pops += slice_pops
            del assign_slice, traj_slice, slice_pops, state_slice

        del futures
        return (assignments, trajlabels, pops, statelabels)

    def go(self):
        assert self.data_reader.parent_id_dsspec._h5file is None
        assert self.data_reader.weight_dsspec._h5file is None
        if hasattr(self.dssynth.dsspec, '_h5file'):
            assert self.dssynth.dsspec._h5file is None
        pi = self.progress.indicator
        pi.operation = 'Initializing'
        with pi, self.data_reader, WESTPAH5File(self.output_filename, 'w', creating_program=True) as self.output_file:
            assign = self.binning.mapper.assign

            # We always assign the entire simulation, so that no trajectory appears to start
            # in a transition region that doesn't get initialized in one.
            iter_start = 1
            iter_stop = self.data_reader.current_iteration

            h5io.stamp_iter_range(self.output_file, iter_start, iter_stop)

            nbins = self.binning.mapper.nbins
            self.output_file.attrs['nbins'] = nbins

            state_map = np.empty((self.binning.mapper.nbins + 1,), index_dtype)
            state_map[:] = 0  # state_id == nstates => unknown state

            # Recursive mappers produce a generator rather than a list of labels
            # so consume the entire generator into a list
            labels = [np.bytes_(label) for label in self.binning.mapper.labels]

            self.output_file.create_dataset('bin_labels', data=labels, compression=9)

            if self.states:
                nstates = len(self.states)
                state_map[:] = nstates  # state_id == nstates => unknown state
                state_labels = [np.bytes_(state['label']) for state in self.states]

                for istate, sdict in enumerate(self.states):
                    assert state_labels[istate] == np.bytes_(sdict['label'])  # sanity check
                    state_assignments = assign(sdict['coords'])
                    for assignment in state_assignments:
                        state_map[assignment] = istate
                self.output_file.create_dataset('state_map', data=state_map, compression=9, shuffle=True)
                self.output_file['state_labels'] = state_labels  # + ['(unknown)']
            else:
                nstates = 0
            self.output_file.attrs['nstates'] = nstates
            # Stamp if this has been subsampled.
            self.output_file.attrs['subsampled'] = self.subsample

            iter_count = iter_stop - iter_start
            nsegs = np.empty((iter_count,), seg_id_dtype)
            npts = np.empty((iter_count,), seg_id_dtype)

            # scan for largest number of segments and largest number of points
            pi.new_operation('Scanning for segment and point counts', iter_stop - iter_start)
            for iiter, n_iter in enumerate(range(iter_start, iter_stop)):
                iter_group = self.data_reader.get_iter_group(n_iter)
                nsegs[iiter], npts[iiter] = iter_group['pcoord'].shape[0:2]
                pi.progress += 1
                del iter_group

            pi.new_operation('Preparing output')

            # create datasets
            self.output_file.create_dataset('nsegs', data=nsegs, shuffle=True, compression=9)
            self.output_file.create_dataset('npts', data=npts, shuffle=True, compression=9)

            max_nsegs = nsegs.max()
            max_npts = npts.max()

            assignments_shape = (iter_count, max_nsegs, max_npts)
            assignments_dtype = np.min_scalar_type(nbins)
            assignments_ds = self.output_file.create_dataset(
                'assignments',
                dtype=assignments_dtype,
                shape=assignments_shape,
                compression=4,
                shuffle=True,
                chunks=h5io.calc_chunksize(assignments_shape, assignments_dtype),
                fillvalue=nbins,
            )
            if self.states:
                trajlabel_dtype = np.min_scalar_type(nstates)
                trajlabels_ds = self.output_file.create_dataset(
                    'trajlabels',
                    dtype=trajlabel_dtype,
                    shape=assignments_shape,
                    compression=4,
                    shuffle=True,
                    chunks=h5io.calc_chunksize(assignments_shape, trajlabel_dtype),
                    fillvalue=nstates,
                )
                statelabels_ds = self.output_file.create_dataset(
                    'statelabels',
                    dtype=trajlabel_dtype,
                    shape=assignments_shape,
                    compression=4,
                    shuffle=True,
                    chunks=h5io.calc_chunksize(assignments_shape, trajlabel_dtype),
                    fillvalue=nstates,
                )

            pops_shape = (iter_count, nstates + 1, nbins + 1)
            pops_ds = self.output_file.create_dataset(
                'labeled_populations',
                dtype=weight_dtype,
                shape=pops_shape,
                compression=4,
                shuffle=True,
                chunks=h5io.calc_chunksize(pops_shape, weight_dtype),
            )
            h5io.label_axes(pops_ds, [np.bytes_(i) for i in ['iteration', 'state', 'bin']])

            pi.new_operation('Assigning to bins', iter_stop - iter_start)
            last_labels = None  # mapping of seg_id to last macrostate inhabited
            for iiter, n_iter in enumerate(range(iter_start, iter_stop)):
                # get iteration info in this block

                if iiter == 0:
                    last_labels = np.empty((nsegs[iiter],), index_dtype)
                    last_labels[:] = nstates  # unknown state

                # Slices this iteration into n_workers groups of segments, submits them to wm, splices results back together
                assignments, trajlabels, pops, statelabels = self.assign_iteration(n_iter, nstates, nbins, state_map, last_labels)

                # Do stuff with this iteration's results

                last_labels = trajlabels[:, -1].copy()
                assignments_ds[iiter, 0 : nsegs[iiter], 0 : npts[iiter]] = assignments
                pops_ds[iiter] = pops
                if self.states:
                    trajlabels_ds[iiter, 0 : nsegs[iiter], 0 : npts[iiter]] = trajlabels
                    statelabels_ds[iiter, 0 : nsegs[iiter], 0 : npts[iiter]] = statelabels

                pi.progress += 1
                del assignments, trajlabels, pops, statelabels

            for dsname in 'assignments', 'npts', 'nsegs', 'labeled_populations', 'statelabels':
                h5io.stamp_iter_range(self.output_file[dsname], iter_start, iter_stop)


def entry_point():
    WAssign().main()


if __name__ == '__main__':
    entry_point()



import logging

import numpy as np

import westpa
from westpa.oldtools.aframe import AnalysisMixin

log = logging.getLogger(__name__)


class BinningMixin(AnalysisMixin):
    '''A mixin for performing binning on WEST data.'''

    def __init__(self):
        super().__init__()

        self.mapper = None
        self.n_bins = None

        self.discard_bin_assignments = False
        self.binning_h5gname = 'binning'
        self.binning_h5group = None
        self.mapper_hash = None

    def add_args(self, parser, upcall=True):
        if upcall:
            try:
                upfunc = super().add_args
            except AttributeError:
                pass
            else:
                upfunc(parser)

        group = parser.add_argument_group('binning options')
        egroup = group.add_mutually_exclusive_group()
        egroup.add_argument(
            '--binexpr',
            '--binbounds',
            dest='binexpr',
            help='''Construct rectilinear bins from BINEXPR. This must be a list of lists of bin boundaries
                            (one list of bin boundaries for each dimension of the progress coordinate), formatted as a Python
                            expression. E.g. "[[0,1,2,4,inf], [-inf,0,inf]]".''',
        )
        group.add_argument(
            '--discard-bin-assignments',
            dest='discard_bin_assignments',
            action='store_true',
            help='''Discard any existing bin assignments stored in the analysis HDF5 file.''',
        )

    def process_args(self, args, upcall=True):
        if args.binexpr:
            westpa.rc.pstatus("Constructing rectilinear bin boundaries from the following expression: '{}'".format(args.binexpr))
            self.mapper = self.mapper_from_expr(args.binexpr)
        else:
            westpa.rc.pstatus('Loading bin boundaries from WEST system')
            system = westpa.rc.get_system_driver()
            self.mapper = system.bin_mapper

        self.n_bins = self.mapper.nbins
        _pdat, self.mapper_hash = self.mapper.pickle_and_hash()
        westpa.rc.pstatus('  {:d} bins'.format(self.n_bins))
        westpa.rc.pstatus('  identity hash {}'.format(self.mapper_hash))

        self.discard_bin_assignments = bool(args.discard_bin_assignments)

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def mapper_from_expr(self, expr):
        from westpa.core.binning import RectilinearBinMapper

        namespace = {'numpy': np, 'np': np, 'inf': float('inf')}

        try:
            return RectilinearBinMapper(eval(expr, namespace))
        except TypeError as e:
            if 'has no len' in str(e):
                raise ValueError('invalid bin boundary specification; you probably forgot to make a list of lists')

    def write_bin_labels(self, dest, header='# bin labels:\n', format='# bin {bin_index:{max_iwidth}d} -- {label!s}\n'):
        '''Print labels for all bins in ``self.mapper`` to ``dest``.  If provided, ``header``
        is printed before any labels.   The ``format`` string specifies how bin labels are to be printed.  Valid entries are:
          * ``bin_index`` -- the zero-based index of the bin
          * ``label`` -- the label, as obtained by ``bin.label``
          * ``max_iwidth`` -- the maximum width (in characters) of the bin index, for pretty alignment
        '''
        dest.write(header or '')
        max_iwidth = len(str(self.mapper.nbins - 1))
        for ibin, label in enumerate(self.mapper.labels):
            dest.write(format.format(bin_index=ibin, label=label, max_iwidth=max_iwidth))

    def require_binning_group(self):
        if self.binning_h5group is None:
            self.binning_h5group = self.anal_h5file.require_group(self.binning_h5gname)
        return self.binning_h5group

    def delete_binning_group(self):
        self.binning_h5group = None
        del self.anal_h5file[self.binning_h5gname]

    def record_data_binhash(self, h5object):
        '''Record the identity hash for self.mapper as an attribute on the given HDF5 object (group or dataset)'''
        h5object.attrs['binhash'] = self.mapper_hash

    def check_data_binhash(self, h5object):
        '''Check whether the recorded bin identity hash on the given HDF5 object matches the identity hash for self.mapper'''
        return h5object.attrs.get('binhash') == self.mapper_hash

    def assign_to_bins(self):
        '''Assign WEST segment data to bins.  Requires the DataReader mixin to be in the inheritance tree'''
        self.require_binning_group()

        n_iters = self.last_iter - self.first_iter + 1
        max_n_segs = self.max_iter_segs_in_range(self.first_iter, self.last_iter)
        pcoord_len = self.get_pcoord_len(self.first_iter)

        assignments = np.zeros((n_iters, max_n_segs, pcoord_len), np.min_scalar_type(self.n_bins))
        populations = np.zeros((n_iters, pcoord_len, self.n_bins), np.float64)

        westpa.rc.pstatus('Assigning to bins...')

        for iiter, n_iter in enumerate(range(self.first_iter, self.last_iter + 1)):
            westpa.rc.pstatus('\r  Iteration {:d}'.format(n_iter), end='')
            seg_index = self.get_seg_index(n_iter)
            pcoords = self.get_iter_group(n_iter)['pcoord'][...]
            weights = seg_index['weight']

            for seg_id in range(len(seg_index)):
                assignments[iiter, seg_id, :] = self.mapper.assign(pcoords[seg_id, :, :])

            for it in range(pcoord_len):
                populations[iiter, it, :] = np.bincount(assignments[iiter, : len(seg_index), it], weights, minlength=self.n_bins)

            westpa.rc.pflush()
            del pcoords, weights, seg_index

        assignments_ds = self.binning_h5group.create_dataset('bin_assignments', data=assignments, compression='gzip')
        populations_ds = self.binning_h5group.create_dataset('bin_populations', data=populations, compression='gzip')

        for h5object in (self.binning_h5group, assignments_ds, populations_ds):
            self.record_data_iter_range(h5object)
            self.record_data_iter_step(h5object, 1)
            self.record_data_binhash(h5object)

        westpa.rc.pstatus()

    def require_bin_assignments(self):
        self.require_binning_group()
        do_assign = False
        if self.discard_bin_assignments:
            westpa.rc.pstatus('Discarding existing bin assignments.')
            do_assign = True
        elif 'bin_assignments' not in self.binning_h5group:
            do_assign = True
        elif not self.check_data_iter_range_least(self.binning_h5group):
            westpa.rc.pstatus('Existing bin assignments are for incompatible first/last iterations; deleting assignments.')
            do_assign = True
        elif not self.check_data_binhash(self.binning_h5group):
            westpa.rc.pstatus('Bin definitions have changed; deleting existing bin assignments.')
            do_assign = True

        if do_assign:
            self.delete_binning_group()
            self.assign_to_bins()
        else:
            westpa.rc.pstatus('Using existing bin assignments.')

    def get_bin_assignments(self, first_iter=None, last_iter=None):
        return self.slice_per_iter_data(self.binning_h5group['bin_assignments'], first_iter, last_iter)

    def get_bin_populations(self, first_iter=None, last_iter=None):
        return self.slice_per_iter_data(self.binning_h5group['bin_populations'], first_iter, last_iter)



import unittest
import argparse
import os
import tempfile
import westpa


class TestDataManager(unittest.TestCase):
    test_name = 'DATA_MANAGER'

    def __init__(self, methodName):
        super().__init__(methodName='test_data_manager')

    def setUp(self):
        parser = argparse.ArgumentParser()
        westpa.rc.add_args(parser)

        here = os.path.dirname(__file__)
        # Set SIM_ROOT to refs folder with west.cfg for odld simulation
        os.environ['WEST_SIM_ROOT'] = os.path.join(here, 'refs', 'odld')

        config_file_name = os.path.join(here, 'refs', 'odld', 'west.cfg')
        args = parser.parse_args(['-r={}'.format(config_file_name), "--verbose"])
        westpa.rc.process_args(args)

        self.sim_manager = westpa.rc.get_sim_manager()
        self.test_dir = tempfile.mkdtemp()
        self.hdf5 = os.path.join(self.test_dir, "west.h5")
        # data = westpa.rc.get_sim_manager().we_driver.rc.get_data_manager()
        # data.we_h5filename = self.hdf5

        self.data_manager = westpa.rc.new_data_manager()
        self.data_manager.we_h5filename = self.hdf5
        self.assertEqual(self.data_manager.h5_access_mode, 'r+')
        """ 1. westpa.rc.get_data_manager() is executed and calls the new_data_manager function.
            2. The new_data_manager instantiates a WESTDataManager object, thus calling the
               WESTDataManager.__init__() constructor. In __init__, process_config() is executed.
            3. process_config() will read in the refs/odld/west.cfg.
               That's all that the setup does in it's current state.
            4. The test checks that the defaults that are set in __init__ are
               indeed the defaults and that the data-manager relevant parts specified
               in west.cfg are indeed updated accordingly."""

    def tearDown(self):
        del os.environ['WEST_SIM_ROOT']
        westpa.rc = westpa.core._rc.WESTRC()

    def test_data_manager(self):
        assert self.data_manager.h5_access_mode == 'r+'
        assert self.data_manager.closed
        assert os.path.basename(self.data_manager.we_h5filename) == 'west.h5'
        assert self.data_manager.aux_compression_threshold == 16384
        assert len(self.data_manager.dataset_options) == 2



'''
Tools for Monte Carlo bootstrap error analysis
'''

import math

import numpy as np
from numpy.random import Generator, MT19937


def msort(input_array):
    return np.sort(input_array, axis=0)


def add_mcbs_options(parser):
    '''Add arguments concerning Monte Carlo bootstrap (``confidence`` and ``bssize``) to the given parser'''

    group = parser.add_argument_group('bootstrapping options')
    group.add_argument(
        '--confidence',
        dest='confidence',
        type=float,
        default=0.95,
        help='Construct a confidence interval of width CONFIDENCE (default: 0.95=95%%)',
    )
    group.add_argument(
        '--bssize',
        dest='bssize',
        type=int,
        help='Use a bootstrap of BSSIZE samples to calculate error (default: chosen from confidence)',
    )


def get_bssize(alpha):
    '''Return a bootstrap data set size appropriate for the given confidence level'''
    return int(10 ** (math.ceil(-math.log10(alpha)) + 1))


def bootstrap_ci(estimator, data, alpha, n_sets=None, args=(), kwargs={}, sort=msort, extended_output=False):
    '''Perform a Monte Carlo bootstrap of a (1-alpha) confidence interval for the given ``estimator``.
    Returns (fhat, ci_lower, ci_upper), where fhat is the result of ``estimator(data, *args, **kwargs)``,
    and ``ci_lower`` and ``ci_upper`` are the lower and upper bounds of the surrounding confidence
    interval, calculated by calling ``estimator(syndata, *args, **kwargs)`` on each synthetic data
    set ``syndata``.  If ``n_sets`` is provided, that is the number of synthetic data sets generated,
    otherwise an appropriate size is selected automatically (see ``get_bssize()``).

    ``sort``, if given, is applied to sort the results of calling ``estimator`` on each
    synthetic data set prior to obtaining the confidence interval.

    Individual entries in synthetic data sets are selected by the first index of ``data``, allowing this
    function to be used on arrays of multidimensional data.

    If ``extended_output`` is True (by default not), instead of returning (fhat, lb, ub), this function returns
    (fhat, lb, ub, ub-lb, abs((ub-lb)/fhat), and max(ub-fhat,fhat-lb)) (that is, the estimated value, the
    lower and upper bounds of the confidence interval, the width of the confidence interval, the relative
    width of the confidence interval, and the symmetrized error bar of the confidence interval).'''

    data = np.asanyarray(data)

    fhat = estimator(data, *args, **kwargs)

    try:
        estimator_shape = fhat.shape
    except AttributeError:
        estimator_shape = ()

    try:
        estimator_dtype = fhat.dtype
    except AttributeError:
        estimator_dtype = type(fhat)

    dlen = len(data)
    n_sets = n_sets or get_bssize(alpha)

    f_synth = np.empty((n_sets,) + estimator_shape, dtype=estimator_dtype)

    rng = Generator(MT19937())

    for i in range(0, n_sets):
        indices = rng.integers(dlen, size=(dlen,))
        f_synth[i] = estimator(data[indices], *args, **kwargs)

    f_synth_sorted = sort(f_synth)
    lbi = int(math.floor(n_sets * alpha / 2))
    ubi = int(math.ceil(n_sets * (1 - alpha / 2)))
    lb = f_synth_sorted[lbi]
    ub = f_synth_sorted[ubi]

    try:
        if extended_output:
            return (fhat, lb, ub, ub - lb, abs((ub - lb) / fhat) if fhat else 0, max(ub - fhat, fhat - lb))
        else:
            return (fhat, lb, ub)
    finally:
        # Do a little explicit memory management
        del f_synth, f_synth_sorted
